import numpy as np
import pandas as pd
from yggdrasil.serialize.tests import test_AsciiTableSerialize as parent


class TestPandasSerialize(parent.TestAsciiTableSerialize):
    r"""Test class for TestPandasSerialize class."""

    _cls = 'PandasSerialize'

    def test_apply_field_names_errors(self):
        r"""Test errors raised by apply_field_names."""
        self.assert_raises(RuntimeError, self.instance.apply_field_names,
                           pd.DataFrame({'x': np.zeros(3), 'y': np.zeros(3)}),
                           field_names=self.instance.get_field_names())
        names = self.testing_options['objects'][0].columns.tolist()
        names[0] = 'invalid'
        self.assert_raises(RuntimeError, self.instance.apply_field_names,
                           pd.DataFrame({k: np.zeros(3) for k in names}),
                           field_names=self.instance.get_field_names())

    def test_func_serialize_errors(self):
        r"""Test errors raised by func_serialize."""
        self.assert_raises(TypeError, self.instance.func_serialize, None)





def install(args=None):
    r"""Check if LPy language will be supported based on if the lpy package
    can be imported from openalea.

    Args:
        args (argparse.Namespace, optional): Arguments parsed from the
            command line. Default to None and is created from sys.argv.

    Returns:
        bool: True if install succeded, False otherwise.

    """
    try:
        from openalea import lpy
        del lpy
        return True
    except ImportError:
        return False


if __name__ == "__main__":
    out = install()
    if out:
        print("LPy interface installed.")
    else:
        raise Exception("Failed to install LPy interface.")



import numpy as np
from yggdrasil import units
from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)


class TestLengthMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for LengthMetaschemaProperty class."""
    
    _mod = 'ArrayMetaschemaProperties'
    _cls = 'LengthMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestLengthMetaschemaProperty, self).__init__(*args, **kwargs)
        nele = 3
        valid = np.zeros(nele, 'float')
        self._valid = [(valid, nele), (units.add_units(valid, 'cm'), nele)]
        self._invalid = [(valid, nele - 1)]
        self._valid_compare = [(nele, nele)]
        self._invalid_compare = [(nele - 1, nele), (nele, nele - 1)]


class TestShapeMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for ShapeMetaschemaProperty class."""
    
    _mod = 'ArrayMetaschemaProperties'
    _cls = 'ShapeMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestShapeMetaschemaProperty, self).__init__(*args, **kwargs)
        nele = (3, 4)
        valid = np.zeros(nele, 'float')
        self._valid = [(valid, nele), (units.add_units(valid, 'cm'), nele)]
        self._invalid = [(valid, (nele[0], nele[1] - 1))]
        self._valid_compare = [(nele, nele),
                               (nele, list(nele))]
        self._invalid_compare = [(nele, nele[::-1]),
                                 (nele, nele[:-1])]



from yggdrasil.interface.YggInterface import YggRpcServer


print("maxMsgSrv(P): Hello!")
rpc = YggRpcServer("maxMsgSrv", "%s", "%s")

while True:
    ret, input = rpc.rpcRecv()
    if not ret:
        break
    print("maxMsgSrv(P): rpcRecv returned %s, input %.10s..." % (ret, input[0]))
    flag = rpc.rpcSend(input[0])
    if not flag:
        raise RuntimeError('maxMsgSrv(P): Error sending reply.')

print("maxMsgSrv(P): Goodbye!")



import numpy as np
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent


class TestJSONBooleanMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for JSONBooleanMetaschemaType class."""
    
    _mod = 'JSONMetaschemaType'
    _cls = 'JSONBooleanMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestJSONBooleanMetaschemaType, self).__init__(*args, **kwargs)
        self._valid_encoded = [{'type': self.import_cls.name}]
        self._valid_decoded = [True, False]
        self._invalid_validate = [None]
        self._invalid_decoded = []
        self._valid_normalize = [('True', True), ('true', True),
                                 ('False', False), ('false', False),
                                 ('hello', 'hello')]


class TestJSONIntegerMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONIntegerMetaschemaType class."""
    
    _cls = 'JSONIntegerMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestJSONIntegerMetaschemaType, self).__init__(*args, **kwargs)
        self._valid_decoded = [int(1), np.int(1)]
        self._invalid_validate = [None]
        self._invalid_decoded = []
        self._valid_normalize = [('1', 1), ('hello', 'hello')]


class TestJSONNullMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONNullMetaschemaType class."""
    
    _cls = 'JSONNullMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestJSONNullMetaschemaType, self).__init__(*args, **kwargs)
        self._valid_decoded = [None]
        self._invalid_validate = ['hello']
        self._invalid_decoded = []
        self._valid_normalize = []


class TestJSONNumberMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONNumberMetaschemaType class."""
    
    _cls = 'JSONNumberMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestJSONNumberMetaschemaType, self).__init__(*args, **kwargs)
        self._valid_decoded = [int(1), np.int(1), float(1), np.float(1)]
        self._invalid_validate = [None]
        self._invalid_decoded = []
        self._valid_normalize = [('1', 1.0), ('1.0', 1.0), ('hello', 'hello')]


class TestJSONStringMetaschemaType(TestJSONBooleanMetaschemaType):
    r"""Test class for JSONStringMetaschemaType class."""
    
    _cls = 'JSONStringMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestJSONStringMetaschemaType, self).__init__(*args, **kwargs)
        self._valid_decoded = ['hello']
        self._invalid_validate = [None]
        self._invalid_decoded = []
        self._valid_normalize = [(1, '1'), (1.0, '1.0'),
                                 ([1, 2, 3], [1, 2, 3])]



import os
from yggdrasil import backwards
from yggdrasil.metaschema.encoder import _use_rapidjson
from yggdrasil.examples.tests import ExampleTstBase


class TestExampleFIO8(ExampleTstBase):
    r"""Test the Formatted I/O lesson 8 example."""

    example_name = 'formatted_io8'

    @property
    def input_files(self):
        r"""Input file."""
        if backwards.PY2:  # pragma: Python 2
            out = [os.path.join(self.yamldir, 'Input', 'input_py2.txt')]
        elif _use_rapidjson:   # pragma: Python 3
            out = [os.path.join(self.yamldir, 'Input', 'input_rj.txt')]
        else:   # pragma: no cover
            out = [os.path.join(self.yamldir, 'Input', 'input.txt')]
        return out

    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.yamldir, 'output.txt')]



import os
import copy
import unittest
from yggdrasil import tools, timing, backwards, platform
from yggdrasil.tests import YggTestClass, assert_raises, long_running


_test_size = 1
_test_count = 1
_test_nrep = 1
_test_lang = 'c'
# On windows, it's possible to not have a C/C++ communication library installed
if 'c' not in timing.get_lang_list():  # pragma: windows
    _test_lang = 'python'
# _test_run = timing.TimedRun(_test_lang, _test_lang)
# _test_run.time_run(_test_count, _test_size, nrep=_test_nrep)
_this_platform = (platform._platform,
                  backwards._python_version,
                  tools.get_default_comm())
_base_environment = {'platform': 'Linux',
                     'python_ver': '2.7',
                     'comm_type': 'zmq'}
_valid_platforms = [('Linux', '2.7', 'zmq'),
                    ('Linux', '2.7', 'ipc'),
                    ('Linux', '3.5', 'zmq'),
                    ('MacOS', '2.7', 'zmq'),
                    ('Windows', '2.7', 'zmq')]
_testfile_json = 'test_run123.json'
_testfile_dat = 'test_run123.dat'


def test_get_source():
    r"""Test getting source file for test."""
    lang_list = timing.get_lang_list()
    dir_list = ['src', 'dst']
    for l in lang_list:
        for d in dir_list:
            fname = timing.get_source(l, d)
            assert(os.path.isfile(fname))


def test_platform_error():
    r"""Test error when test cannot be performed."""
    if platform._is_mac:
        test_platform = 'Linux'
    else:
        test_platform = 'MacOS'
    x = timing.TimedRun(_test_lang, _test_lang, platform=test_platform)
    assert_raises(RuntimeError, x.can_run, raise_error=True)


class TimedRunTestBase(YggTestClass):
    r"""Base test class for the TimedRun class."""

    _mod = 'yggdrasil.timing'
    _cls = 'TimedRun'
    test_name = 'timed_pipe'
    _filename = None
    platform = None
    python_ver = None
    comm_type = None
    dont_use_pyperf = False
    language = _test_lang
    count = 1
    size = 1
    nrep = 1
    max_errors = 5

    @property
    def inst_args(self):
        r"""list: Arguments for creating a class instance."""
        return [self.language, self.language]

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        return {'test_name': self.test_name, 'filename': self._filename,
                'platform': self.platform, 'python_ver': self.python_ver,
                'comm_type': self.comm_type, 'dont_use_pyperf': self.dont_use_pyperf,
                'max_errors': self.max_errors}

    @property
    def time_run_args(self):
        r"""tuple: Arguments for time_run."""
        return (self.count, self.size)

    @property
    def time_run_kwargs(self):
        r"""dict: Keyword arguments for time_run."""
        return {'nrep': self.nrep}

    @property
    def entry_name(self):
        r"""str: Name of the entry for the provided time_run_args."""
        return self.instance.entry_name(*self.time_run_args)

    @property
    def filename(self):
        r"""str: Name of the file where data is stored."""
        return self.instance.filename

    def check_filename(self):
        r"""Raise a unittest.SkipTest error if the filename dosn't exist."""
        if not os.path.isfile(self.filename):  # pragma: debug
            raise unittest.SkipTest("Performance stats file dosn't exist: %s"
                                    % self.filename)

    def get_raw_data(self):
        r"""Get the raw contents of the data file."""
        out = ''
        if os.path.isfile(self.filename):
            with open(self.filename, 'r') as fd:
                out = fd.read()
        return out

    def time_run(self):
        r"""Perform a timed run."""
        self.instance.time_run(*self.time_run_args, **self.time_run_kwargs)


@unittest.skipIf(not tools.check_environ_bool("YGG_TEST_PRODUCTION_RUNS"),
                 'YGG_TEST_PRODUCTION_RUNS not set')
@long_running
class TestTimedRun(TimedRunTestBase):
    r"""Test class for the TimedRun class using existing data."""

    platform = _base_environment['platform']
    python_ver = _base_environment['python_ver']
    comm_type = _base_environment['comm_type']
    language = 'python'

    @property
    def count(self):
        r"""int: Number of messages to use for tests."""
        return self.instance.base_msg_count

    @property
    def size(self):
        r"""int: Size of messages to use for tests."""
        return self.instance.base_msg_size

    def test_json(self):
        r"""Test loading/saving pyperf data as json."""
        self.check_filename()
        old_text = self.get_raw_data()
        x = self.instance.load(as_json=True)
        self.instance.save(x, overwrite=True)
        new_text = self.get_raw_data()
        self.assert_equal(new_text, old_text)

    def test_save(self):
        r"""Test save with/without overwrite."""
        self.check_filename()
        old_text = self.get_raw_data()
        assert_raises(RuntimeError, self.instance.save, self.instance.data)
        self.instance.save(self.instance.data, overwrite=True)
        new_text = self.get_raw_data()
        self.assert_equal(new_text, old_text)

    def test_scaling_count(self):
        r"""Test running scaling with number of messages."""
        self.check_filename()
        kwargs = dict(min_count=self.count, max_count=self.count,
                      nsamples=1, nrep=self.nrep)
        self.instance.scaling_count(self.size, scaling='log', **kwargs)
        self.instance.scaling_count(self.size, scaling='linear',
                                    per_message=True, **kwargs)
        assert_raises(ValueError, self.instance.scaling_count, self.size,
                      scaling='invalid')

    def test_scaling_size(self):
        r"""Test running scaling with size of messages."""
        self.check_filename()
        kwargs = dict(min_size=self.size, max_size=self.size,
                      nsamples=1, nrep=self.nrep)
        self.instance.scaling_size(self.count, scaling='log', **kwargs)
        self.instance.scaling_size(self.count, scaling='linear',
                                   per_message=True, **kwargs)
        assert_raises(ValueError, self.instance.scaling_size, self.count,
                      scaling='invalid')

    def test_plot_scaling_joint(self):
        r"""Test plot_scaling_joint."""
        self.check_filename()
        kwargs = dict(msg_size0=self.size, msg_count0=self.count,
                      msg_size=[self.size], msg_count=[self.count],
                      per_message=True, time_method='bestof')
        self.instance.plot_scaling_joint(**kwargs)

    def test_plot_scaling(self):
        r"""Test plot_scaling corner cases not covered by test_plot_scaling_joint."""
        self.check_filename()
        self.instance.plot_scaling(self.size, [self.count], per_message=True,
                                   time_method='average', yscale='linear')
        self.instance.plot_scaling([self.size], self.count, per_message=False,
                                   time_method='average', yscale='log')

        if False:
            # Test with msg_count on x linear/linear axs
            args = (self.size, [self.count])
            kwargs = {'axs': None, 'nrep': self.nrep,
                      'time_method': 'average', 'per_message': True}
            kwargs['axs'] = self.instance.plot_scaling(*args, **kwargs)
            kwargs['time_method'] = 'bestof'
            kwargs['axs'] = self.instance.plot_scaling(*args, **kwargs)
            # Test with msg_size on x log/log axes
            args = ([self.size], self.count)
            kwargs = {'axs': None, 'nrep': self.nrep, 'time_method': 'average',
                      'xscale': 'log', 'yscale': 'log'}
            kwargs['axs'] = self.instance.plot_scaling(*args, **kwargs)
            kwargs['time_method'] = 'bestof'
            kwargs['axs'] = self.instance.plot_scaling(*args, **kwargs)
        # Errors
        assert_raises(RuntimeError, self.instance.plot_scaling,
                      [self.size], [self.count])
        assert_raises(RuntimeError, self.instance.plot_scaling,
                      self.size, self.count)
        assert_raises(ValueError, self.instance.plot_scaling,
                      [self.size], self.count, nrep=self.nrep,
                      time_method='invalid')

    def test_pyperfjson_to_pandas(self):
        r"""Test pyperfjson_to_pandas."""
        self.check_filename()
        timing.pyperfjson_to_pandas(self.filename)

    def test_fits(self):
        r"""Test fits to scaling on one platform."""
        self.check_filename()
        self.instance.time_per_byte
        self.instance.time_per_message
        self.instance.startup_time

    def test_plot_scalings(self):
        r"""Test plot_scalings corner cases on test platform."""
        self.check_filename()
        kwargs = copy.deepcopy(self.inst_kwargs)
        kwargs.update(msg_size=[self.size], msg_size0=self.size,
                      msg_count=[self.count], msg_count0=self.count,
                      cleanup_plot=True)
        for c in ['comm_type', 'language', 'platform', 'python_ver']:
            ikws = copy.deepcopy(kwargs)
            ikws['compare'] = c
            if c in ikws:
                del ikws[c]
            if c == 'language':
                ikws['per_message'] = True
                ikws['compare_values'] = [self.language]
            timing.plot_scalings(**ikws)
        # Errors
        assert_raises(ValueError, timing.plot_scalings, compare='invalid')
        assert_raises(RuntimeError, timing.plot_scalings, compare='comm_type',
                      comm_type='zmq')

    def test_production_runs(self):
        r"""Test production tests (those used in paper)."""
        self.check_filename()
        # Limit language list for tests
        for c in ['comm_type', 'language', 'platform', 'python_ver']:
            kwargs = copy.deepcopy(self.inst_kwargs)
            kwargs.update(compare=c, cleanup_plot=True, use_paper_values=True)
            if c in kwargs:
                del kwargs[c]
            timing.plot_scalings(**kwargs)
            # Also do MacOS plot w/ Matlab
            if c == 'language':
                kwargs['platform'] = 'MacOS'
                timing.plot_scalings(**kwargs)


@long_running
class TestTimedRunTemp(TimedRunTestBase):
    r"""Test class for the TimedRun class using temporary data."""

    _filename = 'test_run123.json'

    @property
    def description_prefix(self):
        r"""String prefix to prepend docstr test message with."""
        out = super(TestTimedRunTemp, self).description_prefix
        out += ' Temporary'
        return out

    def cleanup_files(self):
        r"""Remove the temporary file if it exists."""
        if os.path.isfile(self.instance.filename):
            os.remove(self.instance.filename)
        if os.path.isfile(self.instance.pyperfscript):  # pragma: debug
            os.remove(self.instance.pyperfscript)

    def setup(self, *args, **kwargs):
        r"""Cleanup the file if it exists and then reload."""
        super(TestTimedRunTemp, self).setup(*args, **kwargs)
        self.cleanup_files()
        self.instance.reload()

    def teardown(self, *args, **kwargs):
        r"""Cleanup temporary files before destroying instance."""
        self.cleanup_files()
        super(TestTimedRunTemp, self).teardown(*args, **kwargs)

    @property
    def time_run_kwargs(self):
        r"""dict: Keyword arguments for time_run."""
        out = super(TestTimedRunTemp, self).time_run_kwargs
        out['overwrite'] = True
        return out

    def test_pyperf_func(self):
        r"""Test pyperf_func."""
        timing.pyperf_func(1, self.instance, self.count, self.size, 0)

    def test_run_overwrite(self):
        r"""Test performing a run twice, the second time with ovewrite."""
        self.time_run()
        # Reload instance to test load for existing file
        self.clear_instance()
        self.time_run()
        self.instance.remove_entry(self.entry_name)
        assert(not self.instance.has_entry(self.entry_name))


@long_running
class TestTimedRunTempNoPyperf(TestTimedRunTemp):
    r"""Test class for the TimedRun class using temporary data without pyperf."""

    _filename = None  # This forces use of standard name with .dat extension
    dont_use_pyperf = True

    @property
    def description_prefix(self):
        r"""String prefix to prepend docstr test message with."""
        out = super(TestTimedRunTempNoPyperf, self).description_prefix
        out += ' (w/o pyperf)'
        return out

    def test_pyperf_func(self):
        r"""Disabled: Test pyperf_func."""
        pass



# Import Python interface
from yggdrasil.interface import YggInput, YggOutput
# Import Python module containing model calculation
from shoot import calc_shoot_mass


# Create input/output channels
ShootGrowthRate = YggInput('shoot_growth_rate')
InitShootMass = YggInput('init_shoot_mass')
TimeStep = YggInput('shoot_time_step')
NextRootMass = YggInput('next_root_mass')
NextShootMass = YggOutput('next_shoot_mass', '%lf\n')

# Receive shoot growth rate
flag, r_s = ShootGrowthRate.recv()
if not flag:
    raise RuntimeError('shoot: Error receiving shoot growth rate.')
print('shoot: Received shoot growth rate: %s' % str(r_s))

# Receive initial shoot mass
flag, S_t = InitShootMass.recv()
if not flag:
    raise RuntimeError('shoot: Error receiving initial shootmass.')
print('shoot: Received initial shoot mass: %s' % str(S_t))

# Receive inital root mass
flag, R_t = NextRootMass.recv()
if not flag:
    raise RuntimeError('shoot: Error receiving initial root mass.')
print('shoot: Received initial root mass: %s' % str(R_t))

# Send initial shoot mass
flag = NextShootMass.send(S_t)
if not flag:
    raise RuntimeError('shoot: Error sending initial shoot mass.')

# Keep advancing until there arn't any new input times
i = 0
while True:
    
    # Receive the time step
    flag, dt = TimeStep.recv()
    if not flag:
        print('shoot: No more time steps.')
        break
    print('shoot: Received next time step: %s' % str(dt))

    # Receive the next root mass
    flag, R_tp1 = NextRootMass.recv()
    if not flag:
        # This raises an error because there must be a root mass for each time step
        raise RuntimeError('shoot: Error receiving root mass for timestep %d.' % (i + 1))
    print('shoot: Received next root mass: %s' % str(R_tp1))

    # Calculate shoot mass
    S_tp1 = calc_shoot_mass(r_s, dt, S_t, R_t, R_tp1)
    print('shoot: Calculated next shoot mass: %s' % str(S_tp1))

    # Output shoot mass
    flag = NextShootMass.send(S_tp1)
    if not flag:
        raise RuntimeError('shoot: Error sending shoot mass for timestep %d.' % (i + 1))

    # Advance masses to next timestep
    S_t = S_tp1
    R_t = R_tp1
    i += 1



from yggdrasil.tests import assert_equal, assert_raises
from yggdrasil import backwards


def test_assert_bytes():
    r"""Ensure that the proper byte types are identified."""
    valid = [b'hello']
    if backwards.PY2:  # pragma: Python 2
        valid += ['hello']
        invalid = [u'hello']
    else:  # pragma: Python 3
        valid += []
        invalid = ['hello', u'hello']
    for v in valid:
        backwards.assert_bytes(v)
    for v in invalid:
        assert_raises(AssertionError, backwards.assert_bytes, v)

        
def test_assert_unicode():
    r"""Ensure that the proper unicode types are identified."""
    valid = [u'hello']
    if backwards.PY2:  # pragma: Python 2
        valid += []
        invalid = ['hello', b'hello']
    else:  # pragma: Python 3
        valid += ['hello']
        invalid = [b'hello']
    for v in valid:
        backwards.assert_unicode(v)
    for v in invalid:
        assert_raises(AssertionError, backwards.assert_unicode, v)
        

def test_assert_str():
    r"""Ensure that the proper str types are identified."""
    valid = ['hello']
    if backwards.PY2:  # pragma: Python 2
        valid += [b'hello']
        invalid = [u'hello']
    else:  # pragma: Python 3
        valid += [u'hello']
        invalid = [b'hello']
    for v in valid:
        backwards.assert_str(v)
    for v in invalid:
        assert_raises(AssertionError, backwards.assert_str, v)


def test_recurse_conv():
    r"""Test error on incorrect type for recurse_conv."""
    assert_raises(TypeError, backwards.recurse_conv, None, backwards.as_unicode)
        

def test_as_unicode():
    r"""Test as_unicode."""
    # Just strings
    res = u'hello'
    vals = ['hello', b'hello', u'hello', bytearray('hello', 'utf-8')]
    for v in vals:
        backwards.assert_unicode(res)
        assert_equal(backwards.as_unicode(v), res)
    assert_raises(TypeError, backwards.as_unicode, 1)
    # Recursive
    valpha = 'abcdef'
    ralpha = [backwards.as_unicode(a) for a in valpha]
    vals_rec = [vals, tuple(vals), {k: v for k, v in zip(valpha, vals)}]
    resl = [res for v in vals]
    resl_rec = [resl, tuple(resl), {k: v for k, v in zip(ralpha, resl)}]
    for v, r in zip(vals_rec, resl_rec):
        assert_equal(backwards.as_unicode(v, recurse=True), r)
    # Integer
    assert_equal(backwards.as_unicode(int(1), convert_types=(int,)), u'1')
    assert_equal(backwards.as_unicode(int(1), allow_pass=True), int(1))
    

def test_as_bytes():
    r"""Test as_bytes."""
    # Just strings
    res = b'hello'
    vals = ['hello', b'hello', u'hello', bytearray('hello', 'utf-8')]
    for v in vals:
        backwards.assert_bytes(res)
        assert_equal(backwards.as_bytes(v), res)
    assert_raises(TypeError, backwards.as_bytes, 1)
    # Recursive
    valpha = 'abcdef'
    ralpha = [backwards.as_bytes(a) for a in valpha]
    vals_rec = [vals, tuple(vals), {k: v for k, v in zip(valpha, vals)}]
    resl = [res for v in vals]
    resl_rec = [resl, tuple(resl), {k: v for k, v in zip(ralpha, resl)}]
    for v, r in zip(vals_rec, resl_rec):
        assert_equal(backwards.as_bytes(v, recurse=True), r)
    # Integer
    assert_equal(backwards.as_bytes(int(1), convert_types=(int,)), b'1')
    assert_equal(backwards.as_bytes(int(1), allow_pass=True), int(1))


def test_as_str():
    r"""Test as_str."""
    # Just strings
    res = 'hello'
    vals = ['hello', b'hello', u'hello', bytearray('hello', 'utf-8')]
    for v in vals:
        backwards.assert_str(res)
        assert_equal(backwards.as_str(v), res)
    assert_raises(TypeError, backwards.as_str, 1)
    # Recursive
    ralpha = 'abcdef'
    valpha = [backwards.as_bytes(a) for a in ralpha]
    vals_rec = [vals, tuple(vals), {k: v for k, v in zip(valpha, vals)}]
    resl = [res for v in vals]
    resl_rec = [resl, tuple(resl), {k: v for k, v in zip(ralpha, resl)}]
    for v, r in zip(vals_rec, resl_rec):
        assert_equal(backwards.as_str(v, recurse=True), r)
    # Integer
    assert_equal(backwards.as_str(int(1), convert_types=(int,)), '1')
    assert_equal(backwards.as_str(int(1), allow_pass=True), int(1))
    
        
def test_match_stype():
    r"""Test string type matching."""
    slist = ['hello', b'hello', u'hello', bytearray('hello', 'utf-8')]
    for s1 in slist:
        for s2 in slist:
            assert_equal(backwards.match_stype(s1, s2), s1)
    assert_raises(TypeError, backwards.match_stype, 1, 'hello')


def test_format_bytes():
    r"""Test formating of bytes string."""
    s0 = "%s, %s"
    ans = "one, one"
    arg0 = "one"
    args = (backwards.as_bytes(arg0), backwards.as_unicode(arg0))
    for cvt in [backwards.as_bytes, backwards.as_unicode]:
        res = backwards.format_bytes(cvt(s0), args)
        assert_equal(res, cvt(ans))
            

def test_encode_escape():
    r"""Test escape encoding."""
    s = 'hello\nhello'
    ans = 'hello\\nhello'
    assert_equal(backwards.encode_escape(s), ans)

    
def test_decode_escape():
    r"""Test esscape decoding."""
    s = 'hello\\nhello'
    ans = 'hello\nhello'
    assert_equal(backwards.decode_escape(s), ans)



import sys
from openalea.lpy import Lsystem
from openalea.plantgl.all import Tesselator
from yggdrasil.interface.YggInterface import (
    YggInput, YggPlyOutput, YggObjOutput, YggAsciiArrayOutput)
from yggdrasil.serialize.PlySerialize import PlyDict
from yggdrasil.serialize.ObjSerialize import ObjDict


# Parse input
error_code = 0
usage = """Run an LPy model from an lpy input file
  usage: {} FILE [output method]."""
if len(sys.argv) not in [2, 3]:
    print(usage.format(__name__))
fname = sys.argv[1]
if len(sys.argv) == 2:
    out_meth = 'ply'
else:
    out_meth = sys.argv[2]

# Connect to I/O channels
in1 = YggInput('LPy_time')
if out_meth == 'ply':
    out = YggPlyOutput('LPy_mesh')
elif out_meth == 'obj':
    out = YggObjOutput('LPy_mesh')
else:
    out = YggAsciiArrayOutput('LPy_mesh', '%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n')

# Create lsystem & discretizer
lsys = Lsystem(fname)
d = Tesselator()

# Continue looping until no more input times
# TODO: Automated unit conversion
mm_to_cm = 0.1
flag = True
while (flag):

    # Receive next step
    flag, result = in1.recv()
    if not flag:
        print("LPy: End of input.")
        break
    niter = result[0]
    print("LPy: Received request for step %d" % niter)

    # Iterate over tree, rewriting rules
    tree = lsys.axiom
    for i in range(niter):
        tree = lsys.iterate(tree, 1)

    # Discretize the scene
    scene = lsys.sceneInterpretation(tree)
    ply_dict = PlyDict.from_scene(scene, d=d, conversion=mm_to_cm)
    mins, maxs = ply_dict.bounds

    # Send output as just verts in each face
    if out_meth == 'ply':
        flag = out.send(ply_dict)
        nface = ply_dict.nface
    elif out_meth == 'obj':
        obj_dict = ObjDict.from_ply(ply_dict)
        flag = out.send(obj_dict)
        nface = ply_dict.nface
    else:
        mesh = ply_dict.mesh
        flag = out.send(mesh)
        nface = len(mesh)
    if not flag:
        print('LPy: Failed to send mesh.')
        error_code = -1
        break
    print('LPy: Sent mesh with %d faces.' % nface)
    print('LPy: \tmins: %f %f %f' % (mins[0], mins[1], mins[2]))
    print('LPy: \tmaxs: %f %f %f' % (maxs[0], maxs[1], maxs[2]))


print('LPy: Exiting with code %d' % error_code)
sys.exit(error_code)



import os
import copy
import shutil
import tempfile
import numpy as np
import unittest
from yggdrasil.metaschema.datatypes.tests import (
    test_JSONObjectMetaschemaType as parent)
from yggdrasil.metaschema.datatypes import PlyMetaschemaType
from yggdrasil.tests import YggTestClassInfo, assert_raises, assert_equal
from yggdrasil.drivers.LPyModelDriver import LPyModelDriver


vcoords = np.array([[0, 0, 0, 0, 1, 1, 1, 1],
                    [0, 0, 1, 1, 0, 0, 1, 1],
                    [0, 1, 1, 0, 0, 1, 1, 0]], 'float32').T
vcolors = np.array([[255, 255, 255, 255, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 255, 255, 255, 255]], 'uint8').T
eindexs = np.array([[0, 1, 2, 3, 2],
                    [1, 2, 3, 0, 0]], 'int32')
ecolors = np.array([[255, 255, 255, 255, 0],
                    [255, 255, 255, 255, 0],
                    [255, 255, 255, 255, 0]], 'uint8')
_test_value = {'material': 'fake_material', 'vertices': [], 'edges': [],
               'faces': [{'vertex_index': [0, 1, 2]},
                         {'vertex_index': [0, 2, 3]},
                         {'vertex_index': [7, 6, 5, 4]},
                         {'vertex_index': [0, 4, 5, 1]},
                         {'vertex_index': [1, 5, 6, 2]},
                         {'vertex_index': [2, 6, 7, 3]},
                         {'vertex_index': [3, 7, 4, 0]}]}
for i in range(len(vcoords)):
    ivert = {}
    for j, k in enumerate('xyz'):
        ivert[k] = vcoords[i, j]
    for j, k in enumerate(['red', 'green', 'blue']):
        ivert[k] = vcolors[i, j]
    _test_value['vertices'].append(ivert)
for i in range(len(eindexs)):
    iedge = {}
    for j, k in enumerate(['vertex1', 'vertex2']):
        iedge[k] = eindexs[i, j]
    for j, k in enumerate(['red', 'green', 'blue']):
        iedge[k] = ecolors[i, j]
    _test_value['edges'].append(iedge)
for f in _test_value['faces']:
    f['vertex_index'] = [np.int32(x) for x in f['vertex_index']]
_test_value_simple = {'vertices': copy.deepcopy(_test_value['vertices']),
                      'faces': [{'vertex_index': [0, 1, 2]},
                                {'vertex_index': [0, 2, 3]}]}
_test_value_int64 = copy.deepcopy(_test_value)
for f in _test_value_int64['faces']:
    f['vertex_index'] = [np.int64(x) for x in f['vertex_index']]


def test_create_schema():
    r"""Test create_schema."""
    assert_raises(RuntimeError, PlyMetaschemaType.create_schema, overwrite=False)
    temp = os.path.join(tempfile.gettempdir(), 'temp_schema')
    old_schema = PlyMetaschemaType.get_schema()
    try:
        shutil.move(PlyMetaschemaType._schema_file, temp)
        new_schema = PlyMetaschemaType.get_schema()
        assert_equal(old_schema, new_schema)
    except BaseException:  # pragma: debug
        shutil.move(temp, PlyMetaschemaType._schema_file)
        raise
    shutil.move(temp, PlyMetaschemaType._schema_file)


def test_translate_ply2fmt_errors():
    r"""Test errors in translate_ply2fmt."""
    assert_raises(ValueError, PlyMetaschemaType.translate_ply2fmt, 'invalid')


def test_translate_ply2py_errors():
    r"""Test errors in translate_ply2py."""
    assert_raises(ValueError, PlyMetaschemaType.translate_ply2py, 'invalid')


def test_translate_py2ply_errors():
    r"""Test errors in translate_py2ply."""
    assert_raises(ValueError, PlyMetaschemaType.translate_py2ply, 'float128')


def test_singular2plural():
    r"""Test conversion from singular element names to plural ones and back."""
    pairs = [('face', 'faces'), ('vertex', 'vertices'),
             ('vertex_index', 'vertex_indices')]
    for s, p in pairs:
        assert_equal(PlyMetaschemaType.singular2plural(s), p)
        assert_equal(PlyMetaschemaType.plural2singular(p), s)
    assert_raises(ValueError, PlyMetaschemaType.plural2singular, 'invalid')


class TestPlyDict(YggTestClassInfo):
    r"""Test for PlyDict class."""
    
    _mod = 'PlyMetaschemaType'
    _cls = 'PlyDict'
    _simple_test = _test_value_simple

    @property
    def mod(self):
        r"""str: Absolute name of module containing class to be tested."""
        return 'yggdrasil.metaschema.datatypes.%s' % self._mod

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        return _test_value

    def test_count_elements(self):
        r"""Test count_elements."""
        self.assert_raises(ValueError, self.instance.count_elements, 'invalid')
        x = self.instance.count_elements('vertices')
        y = self.instance.count_elements('vertex')
        self.assert_equal(x, y)

    def test_mesh(self):
        r"""Test mesh."""
        self.instance.mesh

    def test_merge(self):
        r"""Test merging two ply objects."""
        ply1 = copy.deepcopy(self.instance)
        ply2 = ply1.merge(self.instance)
        ply1.merge([self.instance], no_copy=True)
        self.assert_equal(ply1, ply2)

    def test_append(self):
        r"""Test appending ply objects."""
        basic = self.import_cls(vertices=self.instance['vertices'],
                                faces=[])
        basic.append(self.instance)

    def test_apply_scalar_map(self, _as_obj=False):
        r"""Test applying a scalar colormap."""
        o = copy.deepcopy(self.instance)
        scalar_arr = np.arange(o.count_elements('faces')).astype('float')
        self.assert_raises(NotImplementedError, o.apply_scalar_map,
                           scalar_arr, scale_by_area=True)
        new_faces = []
        if _as_obj:
            for f in o['faces']:
                if len(f) == 3:
                    new_faces.append(f)
        else:
            for f in o['faces']:
                if len(f['vertex_index']) == 3:
                    new_faces.append(f)
        o['faces'] = new_faces
        for scale in ['linear', 'log']:
            o2 = copy.deepcopy(o)
            o1 = o.apply_scalar_map(scalar_arr, scaling=scale, scale_by_area=True)
            o2.apply_scalar_map(scalar_arr, scaling=scale, scale_by_area=True,
                                no_copy=True)
            self.assert_equal(o1, o2)

    @unittest.skipIf(not LPyModelDriver.is_installed(), "LPy library not installed.")
    def test_to_from_scene(self, _as_obj=False):  # pragma: lpy
        r"""Test conversion to/from PlantGL scene."""
        o1 = self.instance
        cls = o1.__class__
        s = o1.to_scene(name='test')
        o2 = cls.from_scene(s)
        # Direct equivalence won't happen unless test is just for simple mesh
        # as faces with more than 3 vertices will be triangulated.
        cls = self.import_cls
        o1 = cls(self._simple_test)
        s = o1.to_scene(name='test')
        o2 = cls.from_scene(s)
        # import pprint
        # print('o2')
        # pprint.pprint(o2)
        # print('o1')
        # pprint.pprint(o1)
        self.assert_equal(o2, o1)

    def test_to_from_dict(self):
        r"""Test transformation to/from dict."""
        x = self.instance.as_dict()
        y = self.import_cls.from_dict(x)
        self.assert_equal(y, self.instance)

    def test_properties(self):
        r"""Test explicit exposure of specific element counts as properties
        against counts based on singular elements."""
        self.instance.bounds
        self.assert_equal(self.instance.nvert, self.instance.count_elements('vertex'))
        self.assert_equal(self.instance.nface, self.instance.count_elements('face'))


class TestPlyMetaschemaType(parent.TestJSONObjectMetaschemaType):
    r"""Test class for PlyMetaschemaType class."""

    _mod = 'PlyMetaschemaType'
    _cls = 'PlyMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestPlyMetaschemaType, self).__init__(*args, **kwargs)
        self._value = _test_value
        self._fulldef = {'type': self.import_cls.name}
        self._typedef = {'type': self.import_cls.name}
        self._valid_encoded = [self._fulldef]
        self._valid_decoded = [self._value,
                               PlyMetaschemaType.PlyDict(**_test_value),
                               {'vertices': [], 'faces': [],
                                'alt_verts': copy.deepcopy(_test_value['vertices'])},
                               _test_value_int64]
        self._invalid_encoded = [{}]
        self._invalid_decoded = [{'vertices': [{k: 0.0 for k in 'xyz'}],
                                  'faces': [{'vertex_index': [0, 1, 2]}]}]
        self._compatible_objects = [(self._value, self._value, None)]
        self._encode_data_kwargs = {'comments': ['Test comment']}

    def test_decode_data_errors(self):
        r"""Test errors in decode_data."""
        self.assert_raises(ValueError, self.import_cls.decode_data, 'hello', None)



from yggdrasil.examples import yamls
from yggdrasil.examples.tests import _test_registry


# Test discovery
for k in yamls.keys():
    if k not in _test_registry:  # pragma: no cover
        print('no test for example', k)



import numpy as np
import pandas as pd
from collections import OrderedDict
from yggdrasil.metaschema.datatypes.ContainerMetaschemaType import (
    ContainerMetaschemaType)


class JSONObjectMetaschemaType(ContainerMetaschemaType):
    r"""Type associated with a map.

    Developer Notes:
        Support for dynamic objects in C/C++ is still under development.

    """

    name = 'object'
    description = 'A container mapping between keys and values.'
    properties = ['properties']
    metadata_properties = ['properties']
    extract_properties = ['properties']
    python_types = (dict, )
    _replaces_existing = True
    
    _container_type = dict
    _json_type = 'object'
    _json_property = 'properties'
    _empty_msg = {}

    @classmethod
    def coerce_type(cls, obj, typedef=None, key_order=None, **kwargs):
        r"""Coerce objects of specific types to match the data type.

        Args:
            obj (object): Object to be coerced.
            typedef (dict, optional): Type defintion that object should be
                coerced to. Defaults to None.
            key_order (list, optional): Order or keys correpsonding to elements in
                a provided list or tuple. Defaults to None.
            **kwargs: Additional keyword arguments are metadata entries that may
                aid in coercing the type.

        Returns:
            object: Coerced object.

        Raises:
            RuntimeError: If obj is a list or tuple, but key_order is not provided.

        """
        from yggdrasil.serialize import pandas2dict, numpy2dict, list2dict
        if isinstance(obj, pd.DataFrame):
            obj = pandas2dict(obj)
        elif isinstance(obj, np.ndarray) and (len(obj.dtype) > 0):
            obj = numpy2dict(obj)
        elif isinstance(obj, (list, tuple)) and (key_order is not None):
            obj = list2dict(obj, names=key_order)
        return obj

    @classmethod
    def _iterate(cls, container):
        r"""Iterate over the contents of the container. Each element returned
        should be a tuple including an index and a value.

        Args:
            container (obj): Object to be iterated over.

        Returns:
            iterator: Iterator over elements in the container.

        """
        for k, v in container.items():
            yield (k, v)

    @classmethod
    def _assign(cls, container, index, value):
        r"""Assign an element in the container to the specified value.

        Args:
            container (obj): Object that element will be assigned to.
            index (obj): Index in the container object where element will be
                assigned.
            value (obj): Value that will be assigned to the element in the
                container object.

        """
        container[index] = value

    @classmethod
    def _has_element(cls, container, index):
        r"""Check to see if an index is in the container.

        Args:
            container (obj): Object that should be checked for index.
            index (obj): Index that should be checked for.

        Returns:
            bool: True if the index is in the container.

        """
        return (index in container)

    @classmethod
    def _encode_data_alias(cls, obj, typedef, func_encode, container_type=None):
        r"""Encode an object's data using a sepcified function.

        Args:
            obj (object): Object to encode.
            typedef (dict): Type definition that should be used to encode the
                object.
            func_encode (callable): Function that should be used to encode
                elements in the container. Defaults to encode_data.
            container_type (type, optional): Type that should be used for the
                output container. Defaults to cls._container_type.

        Returns:
            string: Encoded object.

        """
        if (container_type is None) and isinstance(obj, OrderedDict):
            container_type = OrderedDict
        return super(JSONObjectMetaschemaType, cls)._encode_data_alias(
            obj, typedef, func_encode, container_type=container_type)



import os
from yggdrasil.examples.tests import ExampleTstBase


class TestExampleRPC2(ExampleTstBase):
    r"""Test the rpc_lesson2 example."""
    
    example_name = 'rpc_lesson2'
    niter1 = 3
    niter2 = 5

    @property
    def results(self):
        r"""Result that should be found in output files."""
        result = []
        for niter in [self.niter1, self.niter2]:
            prev1 = 0
            prev2 = 1
            res = ''
            for i in range(niter):
                next = prev1 + prev2
                res += 'fib(%-2d) = %-2d\n' % ((i + 1), next)
                prev2 = prev1
                prev1 = next
            result.append(res)
        return result

    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.tempdir, 'client_output1.txt'),
                os.path.join(self.tempdir, 'client_output2.txt')]



from yggdrasil.tests import assert_raises
from yggdrasil.metaschema.datatypes import conversions


def test_register_conversion_errors():
    r"""Test errors in register_conversion."""
    x = list(conversions._conversion_registry.keys())[0]
    assert_raises(ValueError, conversions.register_conversion, x[0], x[1])


def test_ply2obj():
    r"""Test conversion between ply and obj objects."""
    from yggdrasil.metaschema.datatypes import (
        ObjMetaschemaType, PlyMetaschemaType)
    from yggdrasil.metaschema.datatypes.tests import (
        test_ObjMetaschemaType, test_PlyMetaschemaType)
    obj_class = ObjMetaschemaType.ObjMetaschemaType
    ply_class = PlyMetaschemaType.PlyMetaschemaType
    obj_test = test_ObjMetaschemaType._test_value
    ply_test = test_PlyMetaschemaType._test_value
    # Start at ply
    obj = conversions.ply2obj(ply_test)
    obj_class.validate_instance(obj, {'type': 'obj'})
    ply = conversions.obj2ply(obj)
    ply_class.validate_instance(ply, {'type': 'ply'})
    # Start at obj
    ply = conversions.obj2ply(obj_test)
    ply_class.validate_instance(ply, {'type': 'ply'})
    obj = conversions.ply2obj(ply)
    obj_class.validate_instance(obj, {'type': 'obj'})



import numpy as np
from yggdrasil import units
from yggdrasil.tests import assert_raises, assert_equal
from yggdrasil.metaschema.properties import ScalarMetaschemaProperties
from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)
from yggdrasil.metaschema.datatypes import MetaschemaTypeError


def test_data2dtype_errors():
    r"""Check that error is raised for list, dict, & tuple objects."""
    assert_raises(MetaschemaTypeError, ScalarMetaschemaProperties.data2dtype, [])


def test_definition2dtype_errors():
    r"""Check that error raised if type not specified."""
    assert_raises(KeyError, ScalarMetaschemaProperties.definition2dtype, {})
    assert_raises(RuntimeError, ScalarMetaschemaProperties.definition2dtype,
                  {'type': 'float'})
    assert_equal(ScalarMetaschemaProperties.definition2dtype({'type': 'bytes'}),
                 np.dtype((ScalarMetaschemaProperties._valid_types['bytes'])))


class TestSubtypeMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for SubtypeMetaschemaProperty class."""
    
    _mod = 'ScalarMetaschemaProperties'
    _cls = 'SubtypeMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestSubtypeMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = [(int(1), 'int'), (float(1), 'float')]
        self._invalid = [(int(1), 'float'), (float(1), 'int')]
        self._valid_compare = [('int', 'int')]
        self._invalid_compare = [('float', 'int')]
        self._valid_normalize_schema = [
            ({'subtype': 'float'}, {'subtype': 'float'}),
            ({'units': 'g'}, {'units': 'g', 'subtype': 'float'}),
            ({'units': ''}, {'units': ''})]

    def test_invalid_encode(self):
        r"""Test invalid encode for object dtype."""
        assert_raises(MetaschemaTypeError, self.import_cls.encode, object)


class TestPrecisionMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for PrecisionMetaschemaProperty class."""
    
    _mod = 'ScalarMetaschemaProperties'
    _cls = 'PrecisionMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestPrecisionMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = [(np.int32(1), 32), (np.float16(1), 16)]
        self._invalid = [(np.int32(1), 8), (np.float32(1), 16)]
        self._valid_compare = [(32, 32), (16, 32)]
        self._invalid_compare = [(32, 16)]
        self._valid_normalize_schema = [
            ({'precision': 64}, {'precision': 64}),
            ({'subtype': 'int'}, {'subtype': 'int', 'precision': 64}),
            ({'subtype': 'complex'}, {'subtype': 'complex', 'precision': 128})]


class TestUnitsMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for UnitsMetaschemaProperty class."""
    
    _mod = 'ScalarMetaschemaProperties'
    _cls = 'UnitsMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestUnitsMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = [(1, ''), (units.add_units(1, 'cm'), 'm')]
        self._invalid = [(units.add_units(1, 'cm'), 'kg')]
        self._valid_compare = [('cm', 'cm'), ('cm', 'm'), ('m', 'cm'),
                               ('', 'cm'), ('cm', '')]
        self._invalid_compare = [('cm', 'g')]



from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent


class ValidClass(object):
    def __init__(self, a, b):
        self._input_args = {'a': int(a), 'b': int(b)}

    def __eq__(self, solf):
        if not isinstance(solf, self.__class__):  # pragma: debug
            return False
        return (self._input_args == solf._input_args)


class InvalidClass:  # pragma: no cover
    # Old style class dosn't inherit from object
    pass


class TestInstanceMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for InstanceMetaschemaType class."""

    _mod = 'InstanceMetaschemaType'
    _cls = 'InstanceMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestInstanceMetaschemaType, self).__init__(*args, **kwargs)
        self._typedef.update({'class': ValidClass,
                              'args': {'a': {'type': 'int'},
                                       'b': {'type': 'int'}}})
        # '%s:ValidClass' % __name__
        self._value = ValidClass(0, 1)
        self._valid_encoded = [self.typedef]
        self._valid_decoded = [self._value]
        self._invalid_encoded = [{}]
        self._invalid_decoded = [int(1), 'hello']
        self._invalid_validate += [InvalidClass()]
        self._compatible_objects = [(self._value, self._value, None)]



from __future__ import print_function
import sys
from yggdrasil.interface.YggInterface import YggOutput


def run(args):
    msg_count = int(args[0])
    msg_size = int(args[1])
    print('Hello from Python pipe_src: msg_count = %d, msg_size = %d' % (
        msg_count, msg_size))

    # Ins/outs matching with the the model yaml
    outq = YggOutput('output_pipe')
    print("pipe_src(P): Created I/O channels")

    # Send test message multiple times
    test_msg = b'0' * msg_size
    count = 0
    for i in range(msg_count):
        ret = outq.send(test_msg)
        if not ret:
            raise RuntimeError('pipe_src(P): SEND ERROR ON MSG %d' % i)
        count += 1

    print('Goodbye from Python source. Sent %d messages.' % count)
    

if __name__ == '__main__':
    run(sys.argv[1:])



from yggdrasil.communication import FileComm


class JSONFileComm(FileComm.FileComm):
    r"""Class for handling I/O from/to a JSON file on disk.

    Args:
        name (str): The environment variable where file path is stored.
        **kwargs: Additional keywords arguments are passed to parent class.

    """

    _filetype = 'json'
    _schema_subtype_description = ('The file contains a JSON serialized object.')
    _default_serializer = 'json'



from yggdrasil.drivers.ConnectionDriver import ConnectionDriver
from yggdrasil.drivers.ClientResponseDriver import ClientResponseDriver

# ----
# Client sends resquest to local client output comm
# Client recvs response from local client input comm
# ----
# Client request driver recvs from local client output comm
# Client request driver creates client response driver
# Client request driver sends to server request comm (w/ response comm header)
# ----
# Client response driver recvs from client response comm
# Client response driver sends to local client input comm
# ----
# Server recvs request from local server input comm
# Server sends response to local server output comm
# ----
# Server request driver recvs from server request comm
# Server request driver creates server response driver
# Server request driver sends to local server input comm
# ----
# Server response driver recvs from local server output comm
# Server response driver sends to client response comm
# ----


YGG_CLIENT_INI = b'YGG_BEGIN_CLIENT'
YGG_CLIENT_EOF = b'YGG_END_CLIENT'


class ClientRequestDriver(ConnectionDriver):
    r"""Class for handling client side RPC type communication.

    Args:
        model_request_name (str): The name of the channel used by the client
            model to send requests.
        request_name (str, optional): The name of the channel used to
            send requests to the server request driver. Defaults to
            model_request_name + '_SERVER' if not set.
        comm (str, optional): The comm class that should be used to
            communicate with the server request driver. Defaults to
            tools.get_default_comm().
        comm_address (str, optional): Address for the server request driver.
            Defaults to None and a new address is generated.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        comm (str): The comm class that should be used to communicate with the
            server request driver.
        comm_address (str): Address for the server request driver.
        response_drivers (list): Response drivers created for each request.

    """

    _connection_type = None
    _direction = 'output'

    def __init__(self, model_request_name, request_name=None,
                 comm=None, comm_address=None, **kwargs):
        if request_name is None:
            request_name = model_request_name + '_SERVER'
        # Input communicator
        icomm_kws = kwargs.get('icomm_kws', {})
        icomm_kws['comm'] = None
        icomm_kws['name'] = model_request_name
        kwargs['icomm_kws'] = icomm_kws
        # Output communicator
        ocomm_kws = kwargs.get('ocomm_kws', {})
        ocomm_kws['comm'] = comm
        ocomm_kws['name'] = request_name
        if comm_address is not None:
            ocomm_kws['address'] = comm_address
        ocomm_kws['no_suffix'] = True
        ocomm_kws['is_client'] = True
        ocomm_kws['close_on_eof_send'] = False
        kwargs['ocomm_kws'] = ocomm_kws
        # Parent and attributes
        super(ClientRequestDriver, self).__init__(model_request_name, **kwargs)
        self.env[self.icomm.name] = self.icomm.address
        self.response_drivers = []
        self.comm = comm
        self.comm_address = self.ocomm.opp_address
        self._block_response = False

    @property
    def last_header(self):
        r"""dict: Information contained in the header of the last message
        received from the client model."""
        if self.icomm._last_header is None:
            raise AttributeError("No new requests have been received, so there "
                                 + "does not yet exist information required for "
                                 + "creating a response comm and fowarding the "
                                 + "request.")
        return self.icomm._last_header

    @property
    def request_id(self):
        r"""str: Unique ID for the last message."""
        return self.last_header['request_id']

    @property
    def model_response_address(self):
        r"""str: The address of the channel used by the client model to receive
        responses."""
        return self.last_header['response_address']

    @property
    def request_name(self):
        r"""str: The name of the channel used to send requests to the server
        request driver."""
        return self.ocomm.name
    
    @property
    def request_address(self):
        r"""str: The address of the channel used to send requests to the server
        request driver."""
        return self.ocomm.opp_address

    def close_response_drivers(self):
        r"""Close response driver."""
        # To force response server to connect after response client has stopped
        # self.sleep(0.5)
        with self.lock:
            self.debug("Closing response drivers.")
            self._block_response = True
            for x in self.response_drivers:
                x.terminate()
            self.response_drivers = []

    def close_comm(self):
        r"""Close response drivers."""
        self.close_response_drivers()
        super(ClientRequestDriver, self).close_comm()
            
    def printStatus(self, *args, **kwargs):
        r"""Also print response drivers."""
        super(ClientRequestDriver, self).printStatus(*args, **kwargs)
        for x in self.response_drivers:
            x.printStatus(*args, **kwargs)

    def before_loop(self):
        r"""Send client sign on to server response driver."""
        super(ClientRequestDriver, self).before_loop()
        # self.sleep()  # Help ensure that the server is connected
        self.debug("Sending client sign on")
        super(ClientRequestDriver, self).send_message(YGG_CLIENT_INI,
                                                      header_kwargs={'raw': True})
        self.ocomm._send_serializer = True
        # self.info("%s: before loop complete", self.name)

    def send_message(self, *args, **kwargs):
        r"""Start a response driver for a request message and send message with
        header.

        Args:
            *args: Arguments are passed to parent class send_message.
            **kwargs: Keyword arguments are passed to parent class send_message.

        Returns:
            bool: Success or failure of send.

        """
        if self.ocomm.is_closed:
            return False
        # Start response driver
        is_eof = kwargs.get('is_eof', False)
        if not is_eof:
            with self.lock:
                if (not self.is_comm_open) or self._block_response:  # pragma: debug
                    return False
                drv_args = [self.model_response_address]
                drv_kwargs = dict(comm=self.comm, msg_id=self.request_id,
                                  request_name=self.name)
                self.debug("Creating response comm: address = %s, request_id = %s",
                           self.model_response_address, self.request_id)
                try:
                    response_driver = ClientResponseDriver(*drv_args, **drv_kwargs)
                    self.response_drivers.append(response_driver)
                    response_driver.start()
                    self.debug("Started response comm: address = %s, request_id = %s",
                               self.model_response_address, self.request_id)
                except BaseException:  # pragma: debug
                    self.exception("Could not create/start response driver.")
                    return False
            # Send response address in header
            kwargs.setdefault('header_kwargs', {})
            kwargs['header_kwargs'].setdefault(
                'response_address', response_driver.response_address)
            kwargs['header_kwargs'].setdefault('request_id', self.request_id)
        return super(ClientRequestDriver, self).send_message(*args, **kwargs)



import os
import pprint
import tempfile
from yggdrasil import schema
from yggdrasil.tests import assert_raises, assert_equal


_normalize_objects = [
    ({'models': [{'name': 'modelA',
                  'language': 'c',
                  'args': 'model.c',
                  'outputs': [{'name': 'outputA',
                               'column_names': ['a', 'b'],
                               'column_units': ['cm', 'g']}],
                  'working_dir': os.getcwd()}],
      'connections': [{'inputs': 'outputA',
                       'outputs': 'fileA.txt',
                       'seritype': 'direct',
                       'working_dir': os.getcwd()}]},
     {'models': [{'name': 'modelA',
                  'language': 'c',
                  'args': ['model.c'],
                  'inputs': [], 'outputs': [{'name': 'outputA',
                                             'commtype': 'default',
                                             'datatype': {'type': 'bytes'}}],
                  'working_dir': os.getcwd()}],
      'connections': [{'inputs': [{'name': 'outputA',
                                   'datatype': {'type': 'bytes'},
                                   'commtype': 'default'}],
                       'outputs': [{'name': 'fileA.txt',
                                    'filetype': 'binary',
                                    'working_dir': os.getcwd(),
                                    'serializer': {'seritype': 'direct'},
                                    'field_names': ['a', 'b'],
                                    'field_units': ['cm', 'g']}]}]})]
                                    

def test_SchemaRegistry():
    r"""Test schema registry."""
    assert_raises(ValueError, schema.SchemaRegistry, {})
    x = schema.SchemaRegistry()
    assert_equal(x == 0, False)
    fname = os.path.join(tempfile.gettempdir(), 'temp.yml')
    with open(fname, 'w') as fd:
        fd.write('')
    assert_raises(Exception, x.load, fname)
    os.remove(fname)
    

def test_default_schema():
    r"""Test getting default schema."""
    s = schema.get_schema()
    assert(s is not None)
    schema.clear_schema()
    assert(schema._schema is None)
    s = schema.get_schema()
    assert(s is not None)
    for k in s.keys():
        assert(isinstance(s[k].subtypes, list))
        assert(isinstance(s[k].classes, list))
        for ksub in s[k].classes:
            s[k].get_subtype_properties(ksub)
            s[k].default_subtype


def test_create_schema():
    r"""Test creating new schema."""
    fname = 'test_schema.yml'
    if os.path.isfile(fname):  # pragma: debug
        os.remove(fname)
    # Test saving/loading schema
    s0 = schema.create_schema()
    s0.save(fname)
    assert(s0 is not None)
    assert(os.path.isfile(fname))
    s1 = schema.get_schema(fname)
    assert_equal(s1.schema, s0.schema)
    # assert_equal(s1, s0)
    os.remove(fname)
    # Test getting schema
    s2 = schema.load_schema(fname)
    assert(os.path.isfile(fname))
    assert_equal(s2, s0)
    os.remove(fname)


def test_cdriver2filetype_error():
    r"""Test errors in cdriver2filetype."""
    assert_raises(ValueError, schema.cdriver2filetype, 'invalid')


def test_standardize():
    r"""Test standardize."""
    vals = [(False, ['inputs', 'outputs'], ['_file'],
             {'input': 'inputA', 'output_file': 'outputA'},
             {'inputs': [{'name': 'inputA'}],
              'outputs': [{'name': 'outputA'}]}),
            (True, ['input', 'output'], ['_file'],
             {'inputs': 'inputA', 'output_files': 'outputA'},
             {'input': [{'name': 'inputA'}],
              'output': [{'name': 'outputA'}]})]
    for is_singular, keys, suffixes, x, y in vals:
        schema.standardize(x, keys, suffixes=suffixes, is_singular=is_singular)
        assert_equal(x, y)


def test_normalize():
    r"""Test normalization of legacy formats."""
    s = schema.get_schema()
    for x, y in _normalize_objects:
        a = s.normalize(x, backwards_compat=True)  # , show_errors=True)
        try:
            assert_equal(a, y)
        except BaseException:  # pragma: debug
            print("Unexpected Normalization:\n\nA:")
            pprint.pprint(a)
            print('\nB:')
            pprint.pprint(y)
            raise


def test_cdriver2commtype_error():
    r"""Test error when invalid driver supplied."""
    assert_raises(ValueError, schema.cdriver2commtype, 'invalid')



import os
import copy
import unittest
import jsonschema
from yggdrasil.tests import assert_equal
from yggdrasil.communication import new_comm
from yggdrasil.communication.tests import test_CommBase as parent


def test_wait_for_creation():
    r"""Test FileComm waiting for creation."""
    msg_send = b'Test message\n'
    name = 'temp_file_create.txt'
    kwargs = {'in_temp': True, 'comm': 'FileComm', 'dont_open': True}
    # kwargs = {'wait_for_creation': 5, 'in_temp': True, comm='FileComm'}
    send_instance = new_comm(name, direction='send', **kwargs)
    recv_instance = new_comm(name, direction='recv',
                             wait_for_creation=5.0, **kwargs)
    if os.path.isfile(send_instance.address):
        os.remove(send_instance.address)
    
    def open_and_send(inst, msg):
        inst.open()
        flag = inst.send(msg)
        return flag
    
    send_instance.sched_task(0.5, open_and_send, args=[send_instance, msg_send],
                             store_output=True)
    recv_instance.open()
    T = recv_instance.start_timeout(recv_instance.wait_for_creation)
    while (not T.is_out) and (send_instance.sched_out is None):  # pragma: debug
        recv_instance.sleep()
    recv_instance.stop_timeout()
    assert(send_instance.sched_out)
    flag, msg_recv = recv_instance.recv()
    assert(flag)
    assert_equal(msg_recv, msg_send)
    send_instance.close()
    recv_instance.close()
    recv_instance.remove_file()


class TestFileComm(parent.TestCommBase):
    r"""Test for FileComm communication class."""

    comm = 'FileComm'
    attr_list = (copy.deepcopy(parent.TestCommBase.attr_list)
                 + ['fd', 'read_meth', 'append', 'in_temp',
                    'is_series', 'wait_for_creation', 'serializer',
                    'platform_newline'])
    
    def teardown(self):
        r"""Remove the file."""
        super(TestFileComm, self).teardown()
        self.send_instance.remove_file()

    @property
    def send_inst_kwargs(self):
        r"""dict: Keyword arguments for send instance."""
        out = super(TestFileComm, self).send_inst_kwargs
        out['in_temp'] = True
        return out

    @unittest.skipIf(True, 'File comm')
    def test_send_recv_nolimit(self):
        r"""Disabled: Test send/recv of a large message."""
        pass  # pragma: no cover

    @unittest.skipIf(True, 'File comm')
    def test_work_comm(self):
        r"""Disabled: Test creating/removing a work comm."""
        pass  # pragma: no cover

    def test_invalid_read_meth(self):
        r"""Test raise of error on invalid read_meth."""
        if self.comm == 'FileComm':
            kwargs = self.send_inst_kwargs
            kwargs['read_meth'] = 'invalid'
            kwargs['skip_component_schema_normalization'] = False
            self.assert_raises(jsonschema.ValidationError, new_comm, self.name,
                               **kwargs)

    def test_append(self):
        r"""Test open of file comm with append."""
        send_objects = self.testing_options['send']
        recv_objects = self.testing_options['recv']
        recv_objects_partial = self.testing_options['recv_partial']
        # Write to file
        flag = self.send_instance.send(send_objects[0])
        assert(flag)
        # Create temp file for receving
        recv_kwargs = copy.deepcopy(self.inst_kwargs)
        recv_kwargs['append'] = True
        new_inst_recv = new_comm('partial%s' % self.uuid, **recv_kwargs)
        self.recv_message_list(new_inst_recv, recv_objects_partial[0],
                               break_on_empty=True)
        # Open file in append
        send_kwargs = copy.deepcopy(self.send_inst_kwargs)
        send_kwargs['append'] = True
        new_inst_send = new_comm('append%s' % self.uuid, **send_kwargs)
        for i in range(1, len(send_objects)):
            flag = new_inst_send.send(send_objects[i])
            assert(flag)
            self.recv_message_list(new_inst_recv, recv_objects_partial[i],
                                   break_on_empty=True)
        self.remove_instance(new_inst_send)
        self.remove_instance(new_inst_recv)
        # Read entire contents
        self.recv_message_list(self.recv_instance, recv_objects)
        # Check file contents
        if self.testing_options.get('exact_contents', True):
            with open(self.send_instance.address, 'rb') as fd:
                contents = fd.read()
            self.assert_equal(contents, self.testing_options['contents'])

    def test_series(self):
        r"""Test sending/receiving to/from a series of files."""
        # Set up series
        fname = '%d'.join(os.path.splitext(self.send_instance.address))
        self.send_instance.close()
        self.recv_instance.close()
        self.send_instance.is_series = True
        self.recv_instance.is_series = True
        self.send_instance.address = fname
        self.recv_instance.address = fname
        self.send_instance.open()
        self.recv_instance.open()
        # Send/receive multiple messages
        nmsg = 2
        for i in range(nmsg):
            self.do_send_recv()
        
    def test_remaining_bytes(self):
        r"""Test remaining_bytes."""
        self.assert_equal(self.send_instance.remaining_bytes, 0)
        self.recv_instance.close()
        assert(self.recv_instance.is_closed)
        self.assert_equal(self.recv_instance.remaining_bytes, 0)

    def test_recv_nomsg(self):
        r"""Test recieve when there is no waiting message."""
        flag, msg_recv = self.recv_instance.recv(timeout=self.sleeptime)
        assert(not flag)
        self.assert_equal(msg_recv, self.recv_instance.eof_msg)


class TestFileComm_readline(TestFileComm):
    r"""Test for FileComm communication class with read_meth = 'readline'."""

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        out = super(TestFileComm, self).inst_kwargs
        out['read_meth'] = 'readline'
        return out

    @property
    def testing_options(self):
        r"""dict: Testing options."""
        out = super(TestFileComm_readline, self).testing_options
        out['recv'] = out['send']
        return out



from yggdrasil.metaschema.datatypes.tests import (
    test_JSONObjectMetaschemaType as parent)


class TestSchemaMetaschemaType(parent.TestJSONObjectMetaschemaType):
    r"""Test class for SchemaMetaschemaType class."""

    _mod = 'SchemaMetaschemaType'
    _cls = 'SchemaMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestSchemaMetaschemaType, self).__init__(*args, **kwargs)
        self._value = self._fulldef
        self._fulldef = {'type': 'schema'}
        self._typedef = {'type': 'schema'}
        self._valid_encoded = [self._fulldef]
        self._valid_decoded = [self._value]
        self._invalid_validate = [None]
        self._invalid_decoded = [{}]
        self._invalid_encoded = [{}]
        self._compatible_objects = [(self._value, self._value, None)]
        self._valid_normalize += [('float', {'type': 'float'}),
                                  ({'units': 'g'}, {'units': 'g',
                                                    'type': 'scalar',
                                                    'subtype': 'float',
                                                    'precision': 64}),
                                  ({'title': 'a'}, {'title': 'a'}),
                                  ({'title': 'a', 'units': 'g'},
                                   {'title': 'a', 'units': 'g',
                                    'type': 'scalar', 'subtype': 'float',
                                    'precision': 64}),
                                  ({}, {})]



import os
import gc
import sys
import glob
import logging
import warnings
import argparse
import contextlib
lang_dir = os.path.dirname(__file__)
logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)


def is_path(fpath):
    r"""Determine if a path is valid in a case-sensitive manner.

    Args:
        fpath (str): Full path to be tested.

    Returns:
        bool: True if the path exists, False otherwise.

    """
    pdir, base = os.path.split(fpath)
    return(base in os.listdir(pdir))


def is_file(fname):
    r"""Determine if a file is valid in a case-sensitive manner.

    Args:
        fname (str): Full file path to be tested.

    Returns:
        bool: True if the file exists, False otherwise.

    """
    if not os.path.isfile(fname):
        return False
    return is_path(fname)


def is_dir(fdir):
    r"""Determine if a directory is valid in a case-sensitive manner.

    Args:
        fdir (str): Full directory path to be tested.

    Returns:
        bool: True if the directory exists, False otherwise.

    """
    if not os.path.isdir(fdir):
        return False
    return is_path(fdir)


def get_language_directories():
    r"""Get language directories.

    Returns:
        list: Language directories.

    """
    out = []
    lang_dirs = sorted(glob.glob(os.path.join(lang_dir, '*')))
    for x in lang_dirs:
        if not os.path.isdir(x):
            continue
        base = os.path.basename(x)
        if base not in ['__pycache__', 'tests']:
            out.append(base)
    return out


@contextlib.contextmanager
def import_language_install(language, no_import=False):
    r"""Temporarily import a language installation script.

    Args:
        language (str): Name of language that installation script should be
            imported for.
        no_import (bool, optional): If True, yggdrasil will not be imported.
            Defaults to False.

    Yields:
        module: Language installation module.

    """
    if not is_file(os.path.join(lang_dir, language, 'install.py')):
        if not (no_import or is_dir(os.path.join(lang_dir, language))):
            from yggdrasil.languages import get_language_dir
            fname = os.path.basename(get_language_dir(language))
            with import_language_install(fname, no_import=True) as install:
                yield install
        else:
            yield None
    else:
        try:
            sys.path.insert(0, os.path.join(lang_dir, language))
            import install
            yield install
        finally:
            sys.path.pop(0)
            if install is not None:
                del install
            if 'install' in sys.modules:
                del sys.modules['install']
            gc.collect()


def update_argparser(parser=None, language=None, no_import=None,
                     from_setup=False, arglist=None):
    r"""Update argument parser with language specific arguments.

    Args:
        parser (argparse.ArgumentParser, optional): Existing argument parser
            that should be updated. Default to None and a new argument parser
            will be created.
        language (str, optional): Language that argument parser should be
            updated with options for. Defaults to None and all language options
            will be added.
        no_import (bool, optional): If True, yggdrasil will not be imported.
            Defaults to None and will be set based on sys.argv.
        from_setup (bool, optional): If True, the function is being called from
            setup.py and the positional arguments should not be parsed and
            unrecognized arguments will be ignored. Defaults to False.
        arglist (list, optional): List of arguments to parse. Defaults to
            None and sys.argv is used.

    Returns:
        argparse.ArgumentParser: Argument parser with language specific arguments.

    """
    if arglist is None:
        arglist = sys.argv
    if (no_import is None) and from_setup:
        no_import = True
    all_languages = [x.lower() for x in get_language_directories()]
    if parser is None:
        parser = argparse.ArgumentParser(
            "Run the installation scripts for one or more languages.")
        parser.add_argument('--no-import', action='store_true',
                            help=('Don\'t import the yggdrasil package in '
                                  'calling the installation script.'))
        if not from_setup:
            parser.add_argument('language', nargs='*',
                                choices=(['all'] + all_languages),
                                type=str.lower,
                                default='all',
                                help='One or more languages to install.')
    if ('-h' in arglist) or ('--help' in arglist) or from_setup:
        if no_import is None:
            no_import = False
        if language is None:
            language = get_language_directories()
    else:
        args = parser.parse_known_args(args=arglist)[0]
        if no_import is None:
            no_import = args.no_import
        if language is None:
            if args.language and ('all' not in args.language):
                language = args.language
            else:
                language = all_languages
    if not isinstance(language, (list, tuple)):
        language = [language]
    for ilang in language:
        with import_language_install(ilang, no_import=no_import) as install:
            if hasattr(install, 'update_argparser'):
                parser = install.update_argparser(parser)
    return parser


def install_language(language, results=None, no_import=None, args=None,
                     arglist=None):
    r"""Call install for a specific language.

    Args:
        language (str): Name of language that should be checked.
        results (dict, optional): Dictionary where result (whether or not the
            language is installed) should be logged. Defaults to None and is
            initialized to an empty dict.
        no_import (bool, optional): If True, yggdrasil will not be imported.
            Defaults to None and is set based on the value parsed from
            arglist.
        args (argparse.Namespace, optional): Arguments parsed from the
            command line. Default to None and is created.
        arglist (list, optional): List of arguments to parse. Defaults to
            None and sys.argv is used.

    """
    if args is None:
        args = update_argparser(no_import=no_import,
                                arglist=arglist).parse_args(args=arglist)
    if no_import is None:
        no_import = args.no_import
    if results is None:
        results = {}
    with import_language_install(language, no_import=no_import) as install:
        if install is None:
            logger.info("Nothing to be done for %s" % language)
            name_in_pragmas = language.lower()
            results[name_in_pragmas] = True
            return
        else:
            name_in_pragmas = getattr(install, 'name_in_pragmas', language.lower())
            out = install.install(args=args)
            results[name_in_pragmas] = out
    if not out:
        warnings.warn(("Could not complete installation for {lang}. "
                       "{lang} support will be disabled.").format(lang=language))
    else:
        logger.info("Language %s installed." % language)


def install_all_languages(from_setup=False, arglist=None, **kwargs):
    r"""Call install.py for all languages that have one and return a dictionary
    mapping from language name to the installation state (True if install was
    successful, False otherwise).

    Args:
        from_setup (bool, optional): If True, the function is being called from
            setup.py and the positional arguments should not be parsed and
            unrecognized arguments will be ignored. Defaults to False.
        arglist (list, optional): List of arguments to parse. Defaults to
            None and sys.argv is used.
        **kwargs: Additional keyword arguments are passed to each call to
            install_language.

    Returns:
        dict: Mapping from language name to boolean describing installation
            success.

    """
    if not kwargs.get('args', None):
        kwargs['args'], _ = update_argparser(
            from_setup=True,
            no_import=kwargs.get('no_import', None),
            arglist=arglist).parse_known_args(args=arglist)
    installed_languages = {}
    for ilang in get_language_directories():
        install_language(ilang, installed_languages, **kwargs)
    return installed_languages


if __name__ == "__main__":
    install_all_languages()



import jsonschema
import copy
from yggdrasil.metaschema.datatypes import compare_schema
from yggdrasil.metaschema.datatypes.MetaschemaType import MetaschemaType


def create_fixed_type_class(name, description, base, fixed_properties,
                            target_globals=None, class_name=None, **kwargs):
    r"""Create a fixed class.

    Args:
        name (str); Name of the fixed type.
        description (str): Description of the fixed type.
        base (MetaschemaType): Base class that should be used.
        fixed_properties (dict): Mapping between properties that are fixed
            and the value they are fixed to.
        target_globals (dict, optional): Globals dictionary for module where the
            fixed class should be added. If None, the new class is returned.
            Defaults to None.
        class_name (str, optional): Name that should be given to the class.
            If not provided, defaults to '<name.title()>MetaschemaType'.
        **kwargs: Additional keyword arguments are treated as attributes that
            should be set on the fixed class.

    Returns:
        str, class: The name of the class created if target_globals is provided,
            the created class if target_globals is None.

    """
    if class_name is None:
        class_name = str('%sMetaschemaType' % name.title())
    iattr = {'name': name,
             'description': description,
             'fixed_properties': fixed_properties,
             'specificity': base.specificity + 1}
    iattr.update(kwargs)
    for k in ['properties', 'definition_properties', 'metadata_properties']:
        iattr[k] = copy.deepcopy(getattr(base, k))
        for x in fixed_properties.keys():
            if x == 'type':
                continue
            if x in iattr[k]:
                iattr[k].remove(x)
    new_cls = type(class_name, (FixedMetaschemaType, base, ), iattr)
    if target_globals is not None:
        target_globals[new_cls.__name__] = new_cls
        del new_cls
        return name
    else:
        return new_cls


class FixedMetaschemaType(MetaschemaType):
    r"""Class that should be used to alias another type, but with certain
    properties fixed.

    Arguments:
        **kwargs: All keyword arguments are assumed to be type definition
            properties which will be used to validate serialized/deserialized
            messages.

    Attributes:
        fixed_properties (dict): Type properties of the parent class that are
            fixed and the values they are fixed to.

    """

    is_fixed = True
    fixed_properties = {}

    @classmethod
    def base(cls):
        r"""Get the type that this type is a fixed version of."""
        return cls.__bases__[-1]

    @classmethod
    def typedef_base2fixed(cls, typedef):
        r"""Transform a type definition from the unfixed base type to the fixed
        type alias by removing fixed properties.

        Args:
            typedef (dict): Type definition for the unfixed base type that might
                include properties that are fixed in the base.
        
        Returns:
            dict: Copy of type definition with fixed properties removed.

        """

        out = copy.deepcopy(typedef)
        if out.get('type', None) == cls.base().name:
            out.update(cls.fixed_properties)
            errors = [e for e in compare_schema(typedef, out)]
            if errors:
                error_msg = "Error(s) in comparison with fixed properties.\n"
                for e in errors:
                    error_msg += '\t%s\n' % e
                raise Exception(error_msg)
            for k, v in cls.fixed_properties.items():
                if k in out:
                    del out[k]
            out['type'] = cls.name
        return out

    @classmethod
    def typedef_fixed2base(cls, typedef):
        r"""Transform a type definition from the fixed alias to the unfixed base by
        setting the fixed properties if they are not already present.

        Args:
            typedef (dict): Type definition for the fixed alias type.

        Returns:
            dict: Copy of type definition with fixed properties added.

        """
        out = copy.deepcopy(typedef)
        if out.get('type', None) == cls.name:
            for k, v in cls.fixed_properties.items():
                if (k == 'type') or (k not in cls.base().properties):
                    continue
                if k in out:
                    assert(out[k] == v)
                else:
                    out[k] = v
            out['type'] = cls.base().name
        # if cls.base().is_fixed:
        #     out = cls.base().typedef_fixed2base(out)
        return out

    @classmethod
    def issubtype(cls, t):
        r"""Determine if this type is a subclass of the provided type.

        Args:
            t (str): Type name to check against.

        Returns:
            bool: True if this type is a subtype of the specified type t.

        """
        if super(FixedMetaschemaType, cls).issubtype(t):
            return True
        if cls.base().issubtype(t):
            return True
        return False

    @classmethod
    def updated_fixed_properties(cls, obj):
        r"""Get a version of the fixed properties schema that includes information
        from the object.

        Args:
            obj (object): Object to use to put constraints on the fixed properties
                schema.

        Returns:
            dict: Fixed properties schema with object dependent constraints.

        """
        return copy.deepcopy(cls.fixed_properties)

    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        if not super(FixedMetaschemaType, cls).validate(obj,
                                                        raise_errors=raise_errors):
            return False
        try:
            jsonschema.validate(obj, cls.updated_fixed_properties(obj),
                                cls=cls.validator())
        except (jsonschema.exceptions.ValidationError, AssertionError):
            if raise_errors:
                raise
            return False
        return True

    # This code was unused by any of the test cases, but is kept in case it is
    # needed in the future
    # @classmethod
    # def encode_type(cls, obj, typedef=None, **kwargs):
    #     r"""Encode an object's type definition.

    #     Args:
    #         obj (object): Object to encode.
    #         typedef (dict, optional): Type properties that should be used to
    #             initialize the encoded type definition in certain cases.
    #             Defaults to None and is ignored.
    #         **kwargs: Additional keyword arguments are treated as additional
    #             schema properties.

    #     Raises:
    #         YggTypeError: If the object is not the correct type.

    #     Returns:
    #         dict: Encoded type definition.

    #     """
    #     type_from_base = False
    #     if typedef is None:
    #         typedef = {}
    #     for k, v in cls.fixed_properties.items():
    #         if (k == 'type'):
    #             continue
    #         elif (typedef.get(k, v) != v) or (kwargs.get(k, v) != v):
    #             type_from_base = True
    #             break
    #     if not type_from_base:
    #         return super(FixedMetaschemaType, cls).encode_type(
    #             obj, typedef=typedef, **kwargs)
    #     if isinstance(typedef, dict):
    #         typedef = cls.typedef_fixed2base(typedef)
    #     kwargs = cls.typedef_fixed2base(kwargs)
    #     out = cls.base().encode_type(obj, typedef=typedef, **kwargs)
    #     out = cls.typedef_base2fixed(out)
    #     return out

    @classmethod
    def check_encoded(cls, metadata, typedef=None, raise_errors=False, **kwargs):
        r"""Checks if the metadata for an encoded object matches the type
        definition.

        Args:
            metadata (dict): Meta data to be tested.
            typedef (dict, optional): Type properties that object should
                be tested against. Defaults to None and object may have
                any values for the type properties (so long as they match
                the schema.
            raise_errors (bool, optional): If True, any errors determining that
                encoded object is not of this type will be raised. Defaults to
                False.
            **kwargs: Additional keyword arguments are passed to the parent class.

        Returns:
            bool: True if the metadata matches the type definition, False
                otherwise.

        """
        try:
            out = cls.typedef_base2fixed(metadata)
        except Exception:
            if raise_errors:
                raise
            return False
        return super(FixedMetaschemaType, cls).check_encoded(
            out, typedef=typedef, raise_errors=raise_errors, **kwargs)

    @classmethod
    def extract_typedef(cls, metadata):
        r"""Extract the minimum typedef required for this type from the provided
        metadata.

        Args:
            metadata (dict): Message metadata.

        Returns:
            dict: Encoded type definition with unncessary properties removed.

        """
        out = cls.typedef_base2fixed(metadata)
        out = super(FixedMetaschemaType, cls).extract_typedef(out)
        return out

    def update_typedef(self, **kwargs):
        r"""Update the current typedef with new values.

        Args:
            **kwargs: All keyword arguments are considered to be new type
                definitions. If they are a valid definition property, they
                will be copied to the typedef associated with the instance.

        Returns:
            dict: A dictionary of keyword arguments that were not added to the
                type definition.

        """
        typename = kwargs.get('type', None)
        if typename == self.__class__.base().name:
            kwargs = self.__class__.typedef_base2fixed(kwargs)
        out = super(FixedMetaschemaType, self).update_typedef(**kwargs)
        return out



from yggdrasil import backwards
from yggdrasil.serialize.tests import test_SerializeBase as parent


class TestAsciiMapSerialize(parent.TestSerializeBase):
    r"""Test class for TestAsciiMapSerialize class."""

    _cls = 'AsciiMapSerialize'

    def test_error_delim(self):
        r"""Test error for message with too many delimiters on a line."""
        msg = backwards.as_bytes(self.instance.delimiter.join(
            ['args1', 'val1', 'args2', 'val2']))
        self.assert_raises(ValueError, self.instance.deserialize, msg)

    def test_error_nonstrval(self):
        r"""Test error on serializing dictionary with non-string values."""
        obj = {1: 'here'}
        self.assert_raises(ValueError, self.instance.serialize, obj)

    def test_remove_quotes(self):
        r"""Test deserialization of message with single quotes."""
        send_msg_list = [b"a\t'a_value'\n", b'a\t"a_value"\n',
                         b"a\ta_value\n"]
        recv_msg = {'a': 'a_value'}
        for send_msg in send_msg_list:
            self.assert_equal(self.instance.deserialize(send_msg)[0], recv_msg)



from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)


class TestClassMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for ClassMetaschemaProperty class."""
    
    _mod = 'ClassMetaschemaProperty'
    _cls = 'ClassMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestClassMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = [(int(1), int), (dict(), dict)]
        self._invalid = [(int(1), dict), (dict(), int)]
        self._encode_errors = []
        self._valid_compare = [(int, int), (int, (int, float)),
                               ([int, float], int),
                               ([int, float], (dict, float))]
        self._invalid_compare = [(int, float), (int, (dict, float)),
                                 ((int, dict), float),
                                 ((int, float), (dict, list))]



from yggdrasil.components import import_component


def determine_suffix(no_suffix=False, reverse_names=False,
                     direction='send', **kwargs):
    r"""Determine the suffix that should be used for the comm name.

    Args:
        no_suffix (bool, optional): If True, the suffix will be an empty
            string. Defaults to False.
        reverse_names (bool, optional): If True, the suffix will be
            opposite that indicated by the direction. Defaults to False.
        direction (str, optional): The direction that the comm will
            processing messages. Defaults to 'send'.
        **kwargs: Additional keyword arguments are ignored.

    Returns:
        str: Suffix that will be added to the comm name when producing
            the name of the environment variable where information about
            the comm will be stored.

    Raises:
        ValueError: If the direction is not 'recv' or 'send'.

    """
    if direction not in ['send', 'recv']:
        raise ValueError("Unrecognized message direction: %s" % direction)
    if no_suffix:
        suffix = ''
    else:
        if ((((direction == 'send') and (not reverse_names))
             or ((direction == 'recv') and reverse_names))):
            suffix = '_OUT'
        else:
            suffix = '_IN'
    return suffix


def new_comm(name, comm=None, **kwargs):
    r"""Return a new communicator, creating necessary components for
    communication (queues, sockets, channels, etc.).

    Args:
        name (str): Communicator name.
        comm (str, optional): Name of communicator class.
        **kwargs: Additional keyword arguments are passed to communicator
            class method new_comm.

    Returns:
        Comm: Communicator of given class.

    """
    if isinstance(comm, list):
        if len(comm) == 1:
            kwargs.update(comm[0])
            kwargs.setdefault('name', name)
            return new_comm(**kwargs)
        else:
            kwargs['comm'] = comm
            comm = 'ForkComm'
    comm_cls = import_component('comm', comm)
    if comm in ['DefaultComm', 'default']:
        commtype = kwargs.pop('commtype', 'default')
        assert(commtype == 'default')
    return comm_cls.new_comm(name, **kwargs)


def get_comm(name, **kwargs):
    r"""Return communicator for existing comm components.

    Args:
        name (str): Communicator name.
        **kwargs: Additional keyword arguments are passed to new_comm.

    Returns:
        Comm: Communicator of given class.

    """
    kwargs['dont_create'] = True
    return new_comm(name, **kwargs)
    

def cleanup_comms(comm=None):
    r"""Call cleanup_comms for the appropriate communicator class.

    Args:
        comm (str, optional): Name of communicator class. Defaults to
            tools.get_default_comm() if not provided.

    Returns:
        int: Number of comms closed.

    """
    return import_component('comm', comm).cleanup_comms()


__all__ = ['new_comm', 'get_comm', 'cleanup_comms']



import copy
import numpy as np
from yggdrasil import units
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent
from yggdrasil.metaschema.properties.ScalarMetaschemaProperties import (
    _valid_types)


class TestScalarMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with float."""
    _mod = 'ScalarMetaschemaType'
    _cls = 'ScalarMetaschemaType'
    _prec = 32
    _type = 'float'
    _shape = 1
    _array_contents = None

    def __init__(self, *args, **kwargs):
        super(TestScalarMetaschemaType, self).__init__(*args, **kwargs)
        if not self._explicit:
            self._typedef['subtype'] = self._type
        if self._type == 'bytes':
            dtype = 'S%d' % (self._prec // 8)
        elif self._type == 'unicode':
            dtype = 'U%d' % (self._prec // 32)
        else:
            dtype = '%s%d' % (self._type, self._prec)
        if self._array_contents is None:
            self._array = np.ones(self._shape, dtype)
        else:
            self._array = np.array(self._array_contents, dtype)
        if self._type in ['bytes', 'unicode']:
            dtype_invalid = 'float'
        else:
            dtype_invalid = 'S10'
        self._invalid_array = np.ones(self._shape, dtype_invalid)
        if 'Array' not in self._cls:
            self._value = self._array[0]
            self._invalid_decoded.append(self._array)
            self._invalid_decoded.append(self._invalid_array[0])
        else:
            self._value = self._array
            if self._array.ndim == 1:
                self._invalid_decoded.append(self._array[0])
                self._invalid_decoded.append(np.ones((3, 4), dtype))
            else:
                self._invalid_decoded.append(self._array[0][0])
                self._invalid_decoded.append(self._array[0])
            self._invalid_decoded.append(self._invalid_array)
        self._valid_encoded = [{'type': self.import_cls.name,
                                'precision': self._prec,
                                'units': '',
                                'data': self._value.tobytes()}]
        if not self._explicit:
            self._valid_encoded[0]['subtype'] = self._type
        self._valid_decoded = [self._value]
        if self._type == 'bytes':
            new_dtype = 'S%d' % (self._prec * 2 // 8)
        elif self._type == 'unicode':
            new_dtype = 'U%d' % (self._prec * 2 // 32)
        else:
            new_dtype = '%s%d' % (self._type, self._prec * 2)
        prec_array = self._array.astype(new_dtype)
        if 'Array' not in self._cls:
            self._prec_value = prec_array[0]
        else:
            self._prec_value = prec_array
        self._compatible_objects = [
            (self._value, self._value, None)]
        if not self._explicit:
            self._compatible_objects.append(
                (self._value, self._prec_value, {'subtype': self._type,
                                                 'precision': self._prec * 2}))
        else:
            self._compatible_objects.append(
                (self._value, self._prec_value, {'precision': self._prec * 2}))
        if 'Array' not in self._cls:
            if self._explicit:
                if self._type == 'bytes':
                    self._valid_normalize = [(1, b'1'),
                                             (u'1', b'1')]
                elif self._type == 'unicode':
                    self._valid_normalize = [(1, u'1'),
                                             (b'1', u'1')]
                else:
                    self._valid_normalize = [(str(self._value), self._value),
                                             ('hello', 'hello')]
        if self._explicit and ('Array' not in self._cls):
            self._invalid_encoded.append({'type': 'scalar',
                                          'subtype': 'invalid'})
        self._invalid_validate.append(np.array([None, 1, list()]))

    def test_from_array(self):
        r"""Test getting object from array."""
        test_val = self._value
        test_kws = {}
        if 'units' in self._typedef:
            test_val = units.add_units(test_val, self._typedef['units'])
            test_kws['unit_str'] = self._typedef['units']
        self.assert_equal(self.instance.from_array(self._array, **test_kws),
                          test_val)


# Dynamically create tests for dynamic and explicitly typed scalars
for t in _valid_types.keys():
    iattr_imp = {'_type': t}
    if t == 'complex':
        iattr_imp['_prec'] = 64
    elif t in ('bytes', 'unicode'):
        iattr_imp['_array_contents'] = ['one', 'two', 'three']
        max_len = len(max(iattr_imp['_array_contents'], key=len))
        if t == 'unicode':
            iattr_imp['_prec'] = max_len * 32
        else:
            iattr_imp['_prec'] = max_len * 8
    iattr_exp = copy.deepcopy(iattr_imp)
    iattr_exp['_cls'] = '%sMetaschemaType' % t.title()
    iattr_exp['_explicit'] = True
    cls_imp = type('TestScalarMetaschemaType_%s' % t,
                   (TestScalarMetaschemaType, ), iattr_imp)
    cls_exp = type('Test%s' % iattr_exp['_cls'],
                   (TestScalarMetaschemaType, ), iattr_exp)
    globals()[cls_imp.__name__] = cls_imp
    globals()[cls_exp.__name__] = cls_exp
    del cls_imp, cls_exp


class TestScalarMetaschemaType_prec(TestScalarMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with precision."""

    def __init__(self, *args, **kwargs):
        super(TestScalarMetaschemaType_prec, self).__init__(*args, **kwargs)
        self._typedef['precision'] = self._prec
        self._valid_encoded.append(copy.deepcopy(self._valid_encoded[0]))
        for x in self._invalid_encoded:
            x['precision'] = self._prec / 2  # compatible precision
        # Version with incorrect precision
        self._invalid_encoded.append(copy.deepcopy(self._valid_encoded[0]))
        self._invalid_encoded[-1]['precision'] = self._prec * 2
        self._invalid_decoded.append(self._prec_value)


class TestScalarMetaschemaType_units(TestScalarMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with units."""

    def __init__(self, *args, **kwargs):
        super(TestScalarMetaschemaType_units, self).__init__(*args, **kwargs)
        self._typedef['units'] = 'cm'
        self._valid_encoded.append(copy.deepcopy(self._valid_encoded[0]))
        self._valid_encoded[-1]['units'] = 'cm'
        self._valid_encoded.append(copy.deepcopy(self._valid_encoded[0]))
        self._valid_encoded[-1]['units'] = 'm'
        self._valid_decoded.append(copy.deepcopy(self._valid_decoded[0]))
        self._valid_decoded[-1] = units.add_units(self._valid_decoded[-1], 'm')
        # Version with incorrect units
        self._invalid_encoded.append(copy.deepcopy(self._valid_encoded[0]))
        self._invalid_encoded[-1]['units'] = 's'



import numpy as np
import pandas as pd
import logging
from collections import OrderedDict
from yggdrasil import serialize, backwards, platform
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver
from yggdrasil.drivers.PythonModelDriver import PythonModelDriver
from yggdrasil.drivers.CModelDriver import CModelDriver
from yggdrasil.languages.R import install


logger = logging.getLogger(__name__)


class RModelDriver(InterpretedModelDriver):  # pragma: R
    r"""Base class for running R models.

    Args:
        name (str): Driver name.
        args (str or list): Argument(s) for running the model in matlab.
            Generally, this should be the full path to a Matlab script.
        **kwargs: Additional keyword arguments are passed to parent class's
            __init__ method.

    """
    _schema_subtype_description = ('Model is written in R.')
    language = 'R'
    language_aliases = ['r']
    language_ext = '.R'
    base_languages = ['python']
    default_interpreter = 'Rscript'
    # Dynamically setting the interface library causes circular a import so
    # it is defined here statically. For dynamic import, use the following:
    #     interface_library = PythonModelDriver.interface_library
    interface_library = 'yggdrasil'
    interface_dependencies = install.requirements_from_description()
    # The Batch version causes output to saved to a file rather than directed to
    # stdout so use Rscript instead. For the batch version, use the following:
    #     default_interpreter_flags = ['CMD', 'BATCH' '--vanilla', '--silent']
    default_interpreter_flags = ['--default-packages=methods,utils']
    send_converters = {'table': serialize.consolidate_array}
    type_map = {
        'int': 'integer, bit64::integer64',
        'float': 'double',
        'string': 'character',
        'array': 'list',
        'object': 'list',
        'boolean': 'logical',
        'null': 'NULL',
        'uint': 'integer',
        'complex': 'complex',
        'bytes': 'char (utf-8)',
        'unicode': 'char',
        '1darray': 'list',
        'ndarray': 'list',
        'ply': 'PlyDict',
        'obj': 'ObjDict',
        'schema': 'list'}
    function_param = {
        'import': 'source(\"{filename}\")',
        'interface': 'library(yggdrasil)',
        'input': '{channel} <- YggInterface(\"YggInput\", \"{channel_name}\")',
        'output': '{channel} <- YggInterface(\"YggOutput\", \"{channel_name}\")',
        'table_input': ('{channel} <- YggInterface(\"YggAsciiTableInput\", '
                        '\"{channel_name}\")'),
        'table_output': ('{channel} <- YggInterface(\"YggAsciiTableOutput\", '
                         '\"{channel_name}\", \"{format_str}\")'),
        'recv': 'c({flag_var}, {recv_var}) %<-% {channel}$recv()',
        'send': '{flag_var} <- {channel}$send({send_var})',
        'function_call': '{output_var} <- {function_name}({input_var})',
        'define': '{variable} <- {value}',
        'true': 'TRUE',
        'not': '!',
        'comment': '#',
        'indent': 2 * ' ',
        'quote': '\"',
        'print': 'print(\"{message}\")',
        'fprintf': 'print(sprintf(\"{message}\", {variables}))',
        'error': 'stop(\"{error_msg}\")',
        'block_end': '}',
        'if_begin': 'if({cond}) {{',
        'for_begin': 'for ({iter_var} in {iter_begin}:{iter_end}) {{',
        'while_begin': 'while ({cond}) {{',
        'try_begin': 'tryCatch({',
        'try_except': '}}, error = function({error_var}) {{',
        'try_end': '})',
        'assign': '{name} <- {value}'}

    @classmethod
    def is_library_installed(cls, lib, **kwargs):
        r"""Determine if a dependency is installed.

        Args:
            lib (str): Name of the library that should be checked.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            bool: True if the library is installed, False otherwise.

        """
        if lib not in cls._library_cache:
            try:
                cls.run_executable(['-e', 'library(%s)' % lib])
                cls._library_cache[lib] = True
            except RuntimeError:
                cls._library_cache[lib] = False
        return cls._library_cache[lib]
        
    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to the parent class's
                method.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        kwargs.setdefault('skip_interpreter_flags', True)
        return super(RModelDriver, cls).language_version(**kwargs)

    # @property
    # def debug_flags(self):
    #     r"""list: Flags that should be prepended to an executable command to
    #     enable debugging."""
    #     if self.with_valgrind:
    #         interp = 'R'.join(self.get_interpreter().rsplit('Rscript', 1))
    #         return [interp, '-d', '"valgrind %s"'
    #                 % ' '.join(self.valgrind_flags), '--vanilla', '-f']
    #     return super(RModelDriver, self).debug_flags
        
    def set_env(self):
        r"""Get environment variables that should be set for the model process.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(RModelDriver, self).set_env()
        out['RETICULATE_PYTHON'] = PythonModelDriver.get_interpreter()
        c_linker = CModelDriver.get_tool('linker')
        search_dirs = c_linker.get_search_path(conda_only=True)
        out = CModelDriver.update_ld_library_path(out, paths_to_add=search_dirs,
                                                  add_to_front=True)
        return out
        
    @classmethod
    def comm_atexit(cls, comm):  # pragma: no cover
        r"""Operations performed on comm at exit including draining receive.
        
        Args:
            comm (CommBase): Communication object.

        """
        if comm.direction == 'recv':
            while comm.recv(timeout=0)[0]:
                comm.sleep()
        else:
            comm.send_eof()
        if not getattr(comm, 'dont_backlog', True):
            comm.linger_close()

    @classmethod
    def language2python(cls, robj):
        r"""Prepare an R object for serialization in Python.

        Args:
           robj (object): Python object prepared in R.

        Returns:
            object: Python object in a form that is serialization friendly.

        """
        logger.debug("language2python: %s, %s" % (robj, type(robj)))
        if isinstance(robj, tuple):
            return tuple([cls.language2python(x) for x in robj])
        elif isinstance(robj, list):
            return [cls.language2python(x) for x in robj]
        elif isinstance(robj, dict):
            return {k: cls.language2python(v) for k, v in robj.items()}
        elif isinstance(robj, backwards.string_types):
            return backwards.as_bytes(robj)
        return robj

    @classmethod
    def python2language(cls, pyobj):
        r"""Prepare a python object for transformation in R.

        Args:
            pyobj (object): Python object.

        Returns:
            object: Python object in a form that is R friendly.

        """
        logger.debug("python2language: %s, %s" % (pyobj, type(pyobj)))
        if isinstance(pyobj, tuple):
            return tuple([cls.python2language(x) for x in pyobj])
        elif isinstance(pyobj, list):
            return [cls.python2language(x) for x in pyobj]
        elif isinstance(pyobj, OrderedDict):
            return OrderedDict([(backwards.as_str(k), cls.python2language(v))
                                for k, v in pyobj.items()])
        elif isinstance(pyobj, dict):
            return {backwards.as_str(k): cls.python2language(v)
                    for k, v in pyobj.items()}
        elif isinstance(pyobj, tuple(list(backwards.string_types)
                                     + [np.string_])):
            return backwards.as_str(pyobj)
        elif isinstance(pyobj, pd.DataFrame):
            # R dosn't have int64 and will cast 64bit ints as floats if passed
            # without casting them to int32 first
            for n in pyobj.columns:
                if pyobj[n].dtype == np.dtype('int64'):
                    pyobj[n] = pyobj[n].astype('int32')
                elif ((not backwards.PY2)
                      and (pyobj[n].dtype == np.dtype('object'))
                      and isinstance(pyobj[n][0], backwards.bytes_type)):
                    pyobj[n] = pyobj[n].apply(backwards.as_str)
        return pyobj
    
    @classmethod
    def write_model_wrapper(cls, model_file, model_function, **kwargs):
        r"""Return the lines required to wrap a model function as an integrated
        model.

        Args:
            model_file (str): Full path to the file containing the model
                function's declaration.
            model_function (str): Name of the model function.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Lines of code wrapping the provided model with the necessary
                code to run it as part of an integration.

        """
        if platform._is_win:  # pragma: windows
            model_file = model_file.replace('\\', '/')
        return super(RModelDriver, cls).write_model_wrapper(
            model_file, model_function, **kwargs)



import sys
import numpy as np
from yggdrasil.interface.YggInterface import (
    YggRpcClient, YggOutput)


def main(iterations):
    r"""Function to execute client communication with server that computes
    numbers in the Fibonacci sequence.

    Args:
        iterations (int): The number of Fibonacci numbers to log.

    """

    print('Hello from Python client: iterations = %d ' % iterations)

    # Set up connections matching yaml
    # RPC client-side connection will be $(server_name)_$(client_name)
    rpc = YggRpcClient("server_client", "%d", "%d")
    log = YggOutput("output_log", 'fib(%-2d) = %-2d\n')

    # Iterate over Fibonacci sequence
    for i in range(1, iterations + 1):
        
        # Call the server and receive response
        print('client(Python): Calling fib(%d)' % i)
        ret, result = rpc.call(np.int32(i))
        if not ret:
            raise RuntimeError('client(Python): RPC CALL ERROR')
        fib = result[0]
        print('client(Python): Response fib(%d) = %d' % (i, fib))

        # Log result by sending it to the log connection
        ret = log.send(np.int32(i), fib)
        if not ret:
            raise RuntimeError('client(Python): SEND ERROR')

    print('Goodbye from Python client')

    
if __name__ == '__main__':
    # Take number of iterations from the first argument
    main(int(sys.argv[1]))



import os
import re
import shutil
import logging
from collections import OrderedDict
from yggdrasil import platform, components
from yggdrasil.drivers.CompiledModelDriver import (
    CompilerBase, LinkerBase)
from yggdrasil.drivers.BuildModelDriver import BuildModelDriver
from yggdrasil.drivers import CModelDriver


logger = logging.getLogger(__name__)


class CMakeConfigure(CompilerBase):
    r"""CMake configuration tool."""
    toolname = 'cmake'
    languages = ['cmake', 'c', 'c++']
    is_linker = False
    default_flags = []  # '-H']
    flag_options = OrderedDict([('definitions', '-D%s'),
                                ('sourcedir', ''),  # '-S'
                                ('builddir', '-B%s'),
                                ('configuration', '-DCMAKE_BUILD_TYPE=%s')])
    output_key = None
    compile_only_flag = None
    default_builddir = '.'
    default_archiver = False
    add_libraries = False
    product_files = ['Makefile', 'CMakeCache.txt',
                     'cmake_install.cmake', 'CMakeFiles']
    remove_product_exts = ['CMakeFiles']

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        CompilerBase.before_registration(cls)
        if platform._is_win:  # pragma: windows
            if platform._is_64bit:
                cls.default_flags.append('-DCMAKE_GENERATOR_PLATFORM=x64')
            cls.product_files += ['ALL_BUILD.vcxproj',
                                  'ALL_BUILD.vcxproj.filters',
                                  'Debug', 'Release', 'Win32', 'Win64', 'x64',
                                  'ZERO_CHECK.vcxproj',
                                  'ZERO_CHECK.vcxproj.filters']
            cls.remove_product_exts += ['Debug', 'Release', 'Win32', 'Win64',
                                        'x64', '.dir']
        
    @classmethod
    def append_product(cls, products, src, new, **kwargs):
        r"""Append a product to the specified list along with additional values
        indicated by cls.product_exts.

        Args:
            products (list): List of of existing products that new product
                should be appended to.
            src (list): Input arguments to compilation call that was used to
                generate the output file (usually one or more source files).
            new (str): New product that should be appended to the list.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        """
        kwargs.setdefault('new_dir', new)
        kwargs.setdefault('dont_append_src', True)
        return super(CMakeConfigure, cls).append_product(products, src,
                                                         new, **kwargs)
        
    @classmethod
    def get_output_file(cls, src, dont_link=False, dont_build=None,
                        sourcedir=None, builddir=None, working_dir=None,
                        **kwargs):
        r"""Determine the appropriate output file or directory that will result
        when configuring/building a given source directory.

        Args:
            src (str): Directory containing source files being compiled.
            dont_link (bool, optional): If True, the result assumes that the
                source is just compiled and not linked. If False, the result
                will be the final result after linking.
            dont_build (bool, optional): Alias for dont_link. If not None, this
                keyword overrides the value of dont_link. Defaults to None.
            sourcedir (str, optional): Directory where sources files are located.
                Defaults to None. If None, src will be used to determine the
                value.
            builddir (str, optional): Directory where build tree should be
                created. Defaults to None. If None, sourcedir will be used.
            working_dir (str, optional): Working directory where output file
                should be located. Defaults to None and is ignored.
            **kwargs: Additional keyword arguments are ignored unless dont_link
                is False; then they are passed to get_linker_output_file

        Returns:
            str: Full path to file that will be produced.

        """
        if dont_build is not None:
            dont_link = dont_build
        if isinstance(src, list):
            src = src[0]
        if sourcedir is None:
            if os.path.isfile(src) or os.path.splitext(src)[-1]:
                sourcedir = os.path.dirname(src)
            else:
                sourcedir = src
        if builddir is None:
            builddir = os.path.normpath(os.path.join(sourcedir,
                                                     cls.default_builddir))
        if dont_link:
            out = builddir
            if (not os.path.isabs(out)) and (working_dir is not None):
                out = os.path.normpath(os.path.join(working_dir, out))
        else:
            out = super(CMakeConfigure, cls).get_output_file(
                src, dont_link=dont_link, sourcedir=sourcedir,
                builddir=builddir, working_dir=working_dir, **kwargs)
        return out

    @classmethod
    def get_flags(cls, sourcedir='.', builddir=None, **kwargs):
        r"""Get a list of configuration/generation flags.

        Args:
            sourcedir (str, optional): Directory containing the source files to
                be compiled and the target CMakeLists.txt file. Defaults to '.'
                (the current working directory).
            builddir (str, optional): Directory that will contain the build tree.
                Defaults to '.' (this current working directory).
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Compiler flags.

        Raises:
            RuntimeError: If dont_link is True and the provide outfile and
                builddir keyword arguments point to conflicting paths.
            ValueError: If 'include_dirs' is set ('sourcedir' should be used
                for cmake to specify the location of the source).

        """
        if kwargs.get('dont_link', False):
            if builddir is None:
                outfile = kwargs.get('outfile', None)
                if outfile is None:
                    builddir = os.path.normpath(os.path.join(sourcedir,
                                                             cls.default_builddir))
                else:
                    builddir = outfile
        # Pop target (used for build stage file name, but not for any other
        # part of the build stage)
        kwargs.pop('target', None)
        # Add conda prefix
        # conda_prefix = cls.get_conda_prefix()
        # if conda_prefix:
        #     kwargs.setdefault('definitions', [])
        #     kwargs['definitions'].append('CMAKE_PREFIX_PATH=%s'
        #                                  % os.path.join(conda_prefix, 'lib'))
        #     kwargs['definitions'].append('CMAKE_LIBRARY_PATH=%s'
        #                                  % os.path.join(conda_prefix, 'lib'))
        out = super(CMakeConfigure, cls).get_flags(sourcedir=sourcedir,
                                                   builddir=builddir, **kwargs)
        return out

    @classmethod
    def get_executable_command(cls, args, **kwargs):
        r"""Determine the command required to run the tool using the specified
        arguments and options.

        Args:
            args (list): The arguments that should be passed to the tool. If
                skip_flags is False, these are treated as input files that will
                be used by the tool.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            str: Output to stdout from the command execution.

        """
        assert(len(args) == 1)
        new_args = []
        if args == cls.version_flags:
            new_args = args
        if not kwargs.get('skip_flags', False):
            sourcedir = kwargs.get('sourcedir', args[0])
            if sourcedir != args[0]:  # pragma: debug
                raise RuntimeError(("The argument list "
                                    "contents (='%s') and 'sourcedir' (='%s') "
                                    "keyword specify the same thing, but those "
                                    "provided do not match.")
                                   % (args[0], sourcedir))
            kwargs['sourcedir'] = args[0]
        return super(CMakeConfigure, cls).get_executable_command(new_args, **kwargs)
    
    @classmethod
    def write_wrappers(cls, target=None, sourcedir=None,
                       target_language=None, **kwargs):
        r"""Write any wrappers needed to compile and/or run a model.

        Args:
            **kwargs: Keyword arguments are ignored (only included to
               allow cascade from child classes).

        Returns:
            list: Full paths to any created wrappers.

        """
        if sourcedir is None:  # pragma: debug
            sourcedir = '.'
        out = super(CMakeConfigure, cls).write_wrappers(**kwargs)
        if target is None:
            include_base = 'ygg_cmake.txt'
        else:
            include_base = 'ygg_cmake_%s.txt' % target
        include_file = os.path.join(sourcedir, include_base)
        if (target_language is not None) and ('driver' not in kwargs):
            kwargs['driver'] = components.import_component('model',
                                                           target_language)
        cls.create_include(include_file, target, **kwargs)
        if os.path.isfile(include_file):
            os.remove(include_file)
        return out
        
    @classmethod
    def create_include(cls, fname, target, compile_flags=None, linker_flags=None,
                       driver=None, logging_level=None, configuration='Release',
                       verbose=False, **kwargs):
        r"""Create CMakeList include file with necessary includes,
        definitions, and linker flags.

        Args:
            fname (str): File where the include file should be saved.
            target (str): Target that links should be added to.
            compile_flags (list, optional): Additional compile flags that
                should be set. Defaults to [].
            linker_flags (list, optional): Additional linker flags that
                should be set. Defaults to [].
            driver (CompiledModelDriver, optional): The CompiledModelDriver that
                should be used to get compiler/linker flags. Defaults to
                CModelDriver.
            logging_level (int, optional): Logging level that should be passed
                as a definition to the C compiler. Defaults to None and will be
                ignored.
            configuration (str, optional): Build type/configuration that should
                be built. Defaults to 'Release'. Only used on Windows to
                determin the standard library.
            verbose (bool, optional): If True, the contents of the created file
                are displayed. Defaults to False.
            **kwargs: Additional keyword arguments are ignored.

        Raises:
            ValueError: If a linker or compiler flag cannot be interpreted.

        """
        if target is None:
            target = '${PROJECT_NAME}'
        if compile_flags is None:
            compile_flags = []
        if linker_flags is None:
            linker_flags = []
        if driver is None:  # pragma: debug
            driver = CModelDriver.CModelDriver
        use_library_path = True  # platform._is_win
        library_flags = []
        external_library_flags = []
        internal_library_flags = []
        compile_flags = driver.get_compiler_flags(
            skip_defaults=True, skip_sysroot=True,
            flags=compile_flags, use_library_path=use_library_path,
            dont_link=True, for_model=True, dry_run=True,
            logging_level=logging_level)
        linker_flags = driver.get_linker_flags(
            skip_defaults=True,
            flags=linker_flags, for_model=True, dry_run=True,
            use_library_path='external_library_flags',
            external_library_flags=external_library_flags,
            use_library_path_internal='internal_library_flags',
            internal_library_flags=internal_library_flags,
            skip_library_libs=True, library_flags=library_flags)
        lines = []
        preamble_lines = []
        library_flags += internal_library_flags + external_library_flags
        # Suppress warnings on windows about the security of strcpy etc.
        # and target x64 if the current platform is 64bit
        if platform._is_win:  # pragma: windows
            new_flags = ["/W4", "/EHsc", '/TP', "/nologo",
                         "-D_CRT_SECURE_NO_WARNINGS"]
            if configuration.lower() == 'debug':  # pragma: debug
                new_flags.append("/MTd")
            else:
                new_flags.append("/MT")
            for x in new_flags:
                if x not in compile_flags:
                    compile_flags.append(x)
        # Compilation flags
        for x in compile_flags:
            if x.startswith('-D'):
                preamble_lines.append('ADD_DEFINITIONS(%s)' % x)
            elif x.startswith('-I'):
                xdir = x.split('-I', 1)[-1]
                if platform._is_win:  # pragma: windows
                    xdir = xdir.replace('\\', re.escape('\\'))
                new_dir = 'INCLUDE_DIRECTORIES(%s)' % xdir
                if new_dir not in preamble_lines:
                    preamble_lines.append(new_dir)
            elif x.startswith('-std=c++') or x.startswith('/std=c++'):
                new_def = 'SET(CMAKE_CXX_STANDARD %s)' % x.split('c++')[-1]
                if new_def not in preamble_lines:
                    preamble_lines.append(new_def)
            elif x.startswith('-') or x.startswith('/'):
                new_def = 'ADD_DEFINITIONS(%s)' % x
                if new_def not in preamble_lines:
                    preamble_lines.append(new_def)
            else:
                raise ValueError("Could not parse compiler flag '%s'." % x)
        # Linker flags
        for x in linker_flags:
            if x.startswith('-l'):
                lines.append('TARGET_LINK_LIBRARIES(%s %s)' % (target, x))
            elif x.startswith('-L'):
                libdir = x.split('-L')[-1]
                if platform._is_win:  # pragma: windows
                    libdir = libdir.replace('\\', re.escape('\\'))
                preamble_lines.append('LINK_DIRECTORIES(%s)' % libdir)
            elif x.startswith('/LIBPATH:'):  # pragma: windows
                libdir = x.split('/LIBPATH:')[-1]
                if '"' in libdir:
                    libdir = libdir.split('"')[1]
                if platform._is_win:
                    libdir = libdir.replace('\\', re.escape('\\'))
                preamble_lines.append('LINK_DIRECTORIES(%s)' % libdir)
            elif os.path.isfile(x):
                library_flags.append(x)
            elif x.startswith('-') or x.startswith('/'):
                raise ValueError("Could not parse linker flag '%s'." % x)
            else:
                lines.append('TARGET_LINK_LIBRARIES(%s %s)' % (target, x))
        # Libraries
        for x in library_flags:
            xorig = x
            xd, xf = os.path.split(x)
            xl, xe = os.path.splitext(xf)
            xl = driver.get_tool('linker').libpath2libname(xf)
            if platform._is_win:  # pragma: windows
                x = x.replace('\\', re.escape('\\'))
                xd = xd.replace('\\', re.escape('\\'))
            xn = os.path.splitext(xl)[0]
            new_dir = 'LINK_DIRECTORIES(%s)' % xd
            if new_dir not in preamble_lines:
                preamble_lines.append(new_dir)
            if cls.add_libraries or (xorig in internal_library_flags):
                # if cls.add_libraries:  # pragma: no cover
                # Version adding library
                lines.append('if (NOT TARGET %s)' % xl)
                if xe.lower() in ['.so', '.dll', '.dylib']:  # pragma: no cover
                    lines.append('    ADD_LIBRARY(%s SHARED IMPORTED)' % xl)
                else:
                    lines.append('    ADD_LIBRARY(%s STATIC IMPORTED)' % xl)
                lines += ['    SET_TARGET_PROPERTIES(',
                          '        %s PROPERTIES' % xl,
                          # '        LINKER_LANGUAGE CXX',
                          # '        CXX_STANDARD 11',
                          '        IMPORTED_LOCATION %s)' % x,
                          'endif()',
                          'TARGET_LINK_LIBRARIES(%s %s)' % (target, xl)]
            else:
                # Version finding library
                lines.append('FIND_LIBRARY(%s_LIBRARY NAMES %s %s HINTS %s)'
                             % (xn.upper(), xf, xn, xd))
                lines.append('TARGET_LINK_LIBRARIES(%s ${%s_LIBRARY})'
                             % (target, xn.upper()))
        lines = preamble_lines + lines
        log_msg = 'CMake include file:\n\t' + '\n\t'.join(lines)
        if verbose:
            logger.info(log_msg)
        else:
            logger.debug(log_msg)
        if fname is None:
            return lines
        else:
            if os.path.isfile(fname):
                os.remove(fname)
            with open(fname, 'w') as fd:
                fd.write('\n'.join(lines))

    
class CMakeBuilder(LinkerBase):
    r"""CMake build tool."""
    toolname = 'cmake'
    languages = ['cmake', 'c', 'c++']
    default_flags = []  # '--clean-first']
    output_key = None
    flag_options = OrderedDict([('builddir', {'key': '--build',
                                              'position': 0}),
                                ('target', '--target'),
                                ('configuration', '--config')])
    executable_ext = ''

    @classmethod
    def call(cls, *args, **kwargs):
        r"""Print contents of CMakeCache.txt before raising error."""
        try:
            return super(CMakeBuilder, cls).call(*args, **kwargs)
        except BaseException:  # pragma: debug
            cache = 'CMakeCache.txt'
            if kwargs.get('working_dir', None):
                cache = os.path.join(kwargs['working_dir'], cache)
            if os.path.isfile(cache):
                with open(cache, 'r') as fd:
                    print('CMakeCache.txt:\n%s' % fd.read())
            else:
                print('Cache file does not exist: %s' % cache)
            raise

    @classmethod
    def extract_kwargs(cls, kwargs, **kwargs_ex):
        r"""Extract linker kwargs, leaving behind just compiler kwargs.

        Args:
            kwargs (dict): Keyword arguments passed to the compiler that should
                be sorted into kwargs used by either the compiler or linker or
                both. Keywords that are not used by the compiler will be removed
                from this dictionary.
            **kwargs_ex: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Keyword arguments that should be passed to the linker.

        """
        kwargs_ex['add_kws_both'] = (kwargs.get('add_kws_both', [])
                                     + ['builddir', 'target'])
        return super(CMakeBuilder, cls).extract_kwargs(kwargs, **kwargs_ex)

    @classmethod
    def get_output_file(cls, obj, target=None, builddir=None, **kwargs):
        r"""Determine the appropriate output file that will result when bulding
        a given directory.

        Args:
            obj (str): Directory being built or a file in the directory being
                built.
            target (str, optional): Target that will be used to create the
                output file. Defaults to None. Target is required in order
                to determine the name of the file that will be created.
            builddir (str, optional): Directory where build tree should be
                created. Defaults to None and obj will used (if its a directory)
                or the directory containing obj will be used (if its a file).
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            str: Full path to file that will be produced.

        Raises:
            RuntimeError: If target is None.

        """
        if builddir is None:
            if os.path.isfile(obj) or os.path.splitext(obj)[-1]:
                builddir = os.path.dirname(obj)
            else:
                builddir = obj
        if target is None:
            if os.path.isfile(obj) or os.path.splitext(obj)[-1]:
                target = os.path.splitext(os.path.basename(obj))[0]
            else:
                raise RuntimeError("Target is required.")
        elif target == 'clean':
            return target
        out = super(CMakeBuilder, cls).get_output_file(
            os.path.join(builddir, target), **kwargs)
        return out

    @classmethod
    def get_flags(cls, target=None, builddir=None, **kwargs):
        r"""Get a list of build flags for building a project using cmake.

        Args:
            target (str, optional): Target that should be built. Defaults to
                to None and is ignored.
            builddir (str, optional): Directory containing the build tree.
                Defaults to None and is set based on outfile is provided or
                cls.default_builddir if not.
                Defaults to '.' (which will be the current working directory).
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Linker flags.

        """
        outfile = kwargs.get('outfile', None)
        if outfile is not None:
            if target is None:
                target = os.path.splitext(os.path.basename(outfile))[0]
            if builddir is None:
                builddir = os.path.dirname(outfile)
        if builddir is None:
            builddir = CMakeConfigure.default_builddir
        out = super(CMakeBuilder, cls).get_flags(target=target,
                                                 builddir=builddir, **kwargs)
        return out

    @classmethod
    def get_executable_command(cls, args, **kwargs):
        r"""Determine the command required to run the tool using the specified
        arguments and options.

        Args:
            args (list): The arguments that should be passed to the tool. If
                skip_flags is False, these are treated as input files that will
                be used by the tool.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            str: Output to stdout from the command execution.

        """
        assert(len(args) == 1)
        if not kwargs.get('skip_flags', False):
            builddir = kwargs.get('builddir', args[0])
            if not os.path.isabs(builddir) and os.path.isabs(args[0]):
                builddir = os.path.join(os.path.dirname(args[0]), builddir)
            if builddir != args[0]:  # pragma: debug
                raise RuntimeError(("The argument list "
                                    "contents (='%s') and 'builddir' (='%s') "
                                    "keyword specify the same thing, but those "
                                    "provided do not match.")
                                   % (args[0], builddir))
            kwargs['builddir'] = args[0]
        return super(CMakeBuilder, cls).get_executable_command([], **kwargs)
    

class CMakeModelDriver(BuildModelDriver):
    r"""Class for running cmake compiled drivers. Before running the
    cmake command, the cmake commands for setting the necessary compiler & linker
    flags for the interface's C/C++ library are written to a file called
    'ygg_cmake.txt' that should be included in the CMakeLists.txt file (after
    the target executable has been added).

    Args:
        name (str): Driver name.
        args (str, list): Executable that should be created (cmake target) and
            any arguments for the executable.
        sourcedir (str, optional): Source directory to call cmake on. If not
            provided it is set to working_dir. This should be the directory
            containing the CMakeLists.txt file. It can be relative to
            working_dir or absolute.
        builddir (str, optional): Directory where the build should be saved.
            Defaults to <sourcedir>/build. It can be relative to working_dir
            or absolute.
        target (str, optional): Make target that should be built to create the
            model executable. Defaults to None.
        target_language (str, optional): Language that the target is written in.
            Defaults to None and will be set based on the source files provided.
        configuration (str, optional): Build type/configuration that should be
            built. Defaults to 'Release'.
        **kwargs: Additional keyword arguments are passed to parent class.

    Attributes:
        sourcedir (str): Source directory to call cmake on.
        builddir (str): Directory where the build should be saved.
        target (str): Name of executable that should be created and called.
        add_libraries (bool): If True, interface libraries and dependency
            libraries are added using CMake's ADD_LIBRARY directive. If False,
            interface libraries are found using FIND_LIBRARY.
        target_language (str): Language that the target is written in.
        target_language_driver (ModelDriver): Language driver for the target
            language.
        configuration (str): Build type/configuration that should be built.
            This is only used on Windows.

    Raises:
        RuntimeError: If neither the IPC or ZMQ C libraries are available.

    """

    _schema_subtype_description = ('Model is written in C/C++ and has a '
                                   'CMake build system.')
    _schema_properties = {'sourcedir': {'type': 'string'},
                          'builddir': {'type': 'string'},
                          'target': {'type': 'string'},
                          'target_language': {'type': 'string'},
                          'configuration': {'type': 'string',
                                            'default': 'Release'}}
    language = 'cmake'
    base_languages = ['c', 'c++']
    add_libraries = CMakeConfigure.add_libraries
    sourcedir_as_sourcefile = True

    def parse_arguments(self, args, **kwargs):
        r"""Sort arguments based on their syntax to determine if an argument
        is a source file, compilation flag, or runtime option/flag that should
        be passed to the model executable.

        Args:
            args (list): List of arguments provided.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        """
        if self.target is None:
            self.builddir_base = 'build'
        else:
            self.builddir_base = 'build_%s' % self.target
        self.buildfile_base = 'CMakeLists.txt'
        super(CMakeModelDriver, self).parse_arguments(args, **kwargs)
        self.buildfile_copy = '_orig'.join(os.path.splitext(self.buildfile))
        self.modified_files.append((self.buildfile_copy, self.buildfile))

    def write_wrappers(self, **kwargs):
        r"""Write any wrappers needed to compile and/or run a model.

        Args:
            **kwargs: Keyword arguments are passed to the parent class's method.

        Returns:
            list: Full paths to any created wrappers.

        """
        out = super(CMakeModelDriver, self).write_wrappers(**kwargs)
        # Create cmake files that can be included
        if self.target is None:
            include_base = 'ygg_cmake.txt'
        else:
            include_base = 'ygg_cmake_%s.txt' % self.target
        include_file = os.path.join(self.sourcedir, include_base)
        self.get_tool_instance('compiler').create_include(
            include_file, self.target,
            driver=self.target_language_driver,
            logging_level=self.logger.getEffectiveLevel(),
            configuration=self.configuration)
        assert(os.path.isfile(include_file))
        out.append(include_file)
        # Create copy of cmakelists and modify
        newlines_before = []
        newlines_after = []
        if os.path.isfile(self.buildfile):
            if not os.path.isfile(self.buildfile_copy):
                shutil.copy2(self.buildfile, self.buildfile_copy)
            with open(self.buildfile, 'r') as fd:
                contents = fd.read().splitlines()
            # Prevent error when cross compiling by building static lib as test
            newlines_before.append(
                'set(CMAKE_TRY_COMPILE_TARGET_TYPE "STATIC_LIBRARY")')
            # Add conda prefix as first line so that conda installed C libraries are
            # used
            conda_prefix = self.get_tool_instance('compiler').get_conda_prefix()
            if conda_prefix:
                newlines_before.append(
                    'LINK_DIRECTORIES(%s)' % os.path.join(conda_prefix, 'lib'))
            # Explicitly set Release/Debug directories to builddir on windows
            if platform._is_win:  # pragma: windows
                for artifact in ['runtime', 'library', 'archive']:
                    for conf in ['release', 'debug']:
                        newlines_before.append(
                            'SET( CMAKE_%s_OUTPUT_DIRECTORY_%s '
                            % (artifact.upper(), conf.upper())
                            + '"${OUTPUT_DIRECTORY}")')
            # Add yggdrasil created include if not already in the file
            newlines_after.append(
                'INCLUDE(%s)' % os.path.basename(include_file))
            # Consolidate lines, checking for lines that already exist
            lines = []
            for newline in newlines_before:
                if newline not in contents:
                    lines.append(newline)
            lines += contents
            for newline in newlines_after:
                if newline not in contents:
                    lines.append(newline)
            # Write contents to the build file, check for new lines that may
            # already be included
            log_msg = 'New CMakeLists.txt:\n\t' + '\n\t'.join(lines)
            if kwargs.get('verbose', False):
                logger.info(log_msg)
            else:
                logger.debug(log_msg)
            with open(self.buildfile, 'w') as fd:
                fd.write('\n'.join(lines))
        return out

    def set_target_language(self):
        r"""Set the language of the target being compiled (usually the same
        as the language associated with this driver.

        Returns:
            str: Name of language.

        """
        if self.target_language is None:
            if os.path.isfile(self.buildfile):
                with open(self.buildfile, 'r') as fd:
                    lines = fd.readlines()
                for x in lines:
                    if not x.strip().upper().startswith('ADD_EXECUTABLE'):
                        continue
                    varlist = x.split('(', 1)[-1].rsplit(')', 1)[0].split()
                    if (self.target is None) or (self.target == varlist[0]):
                        try:
                            self.target_language = (
                                self.get_language_for_source(varlist[1:],
                                                             early_exit=True))
                        except ValueError:  # pragma: debug
                            pass
        return super(CMakeModelDriver, self).set_target_language()
        
    def compile_model(self, target=None, **kwargs):
        r"""Compile model executable(s) and appends any products produced by
        the compilation that should be removed after the run is complete.

        Args:
            target (str, optional): Target to build.
            **kwargs: Keyword arguments are passed on to the call_compiler
                method.

        """
        if target is None:
            target = self.target
        if target == 'clean':
            return self.call_linker(self.builddir, target=target, out=target,
                                    overwrite=True, working_dir=self.working_dir,
                                    allow_error=True, **kwargs)
        else:
            default_kwargs = dict(target=target,
                                  sourcedir=self.sourcedir,
                                  builddir=self.builddir,
                                  skip_interface_flags=True)
            default_kwargs['configuration'] = self.configuration
            for k, v in default_kwargs.items():
                kwargs.setdefault(k, v)
            return super(CMakeModelDriver, self).compile_model(**kwargs)

    @classmethod
    def update_compiler_kwargs(cls, **kwargs):
        r"""Update keyword arguments supplied to the compiler get_flags method
        for various options.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Keyword arguments for a get_flags method providing compiler
                flags.

        """
        out = super(CMakeModelDriver, cls).update_compiler_kwargs(**kwargs)
        if CModelDriver._osx_sysroot is not None:
            out.setdefault('definitions', [])
            out['definitions'].append(
                'CMAKE_OSX_SYSROOT=%s' % CModelDriver._osx_sysroot)
            if os.environ.get('MACOSX_DEPLOYMENT_TARGET', None):
                out['definitions'].append(
                    'CMAKE_OSX_DEPLOYMENT_TARGET=%s'
                    % os.environ['MACOSX_DEPLOYMENT_TARGET'])
        return out
    
    def set_env(self, **kwargs):
        r"""Get environment variables that should be set for the model process.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(CMakeModelDriver, self).set_env(**kwargs)
        if hasattr(self.target_language_driver, 'update_ld_library_path'):
            self.target_language_driver.update_ld_library_path(out)
        # C++ may also need the C libraries
        if self.target_language == 'c++':
            out = CModelDriver.CModelDriver.update_ld_library_path(out)
        return out



import copy
import numpy as np
from yggdrasil import serialize
from yggdrasil.tests import assert_equal
from yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType import (
    JSONObjectMetaschemaType)
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent
from yggdrasil.metaschema.datatypes.tests import (
    test_ContainerMetaschemaType as container_utils)


def test_coerce():
    r"""Test serialization of coerced types."""
    typedef = {'type': 'object',
               'properties': {'a': {'type': '1darray',
                                    'subtype': 'float',
                                    'title': 'a',
                                    'precision': 64}}}
    x = JSONObjectMetaschemaType(**typedef)
    key_order = ['a']
    msg_recv = {'a': np.zeros(3, 'float64')}
    msg_send_list = [serialize.dict2numpy(msg_recv, order=key_order),
                     serialize.dict2pandas(msg_recv, order=key_order),
                     serialize.dict2list(msg_recv, order=key_order)]

    def do_send_recv(msg_send):
        msg_seri = x.serialize(msg_send, tyepdef=typedef, key_order=key_order)
        assert_equal(x.deserialize(msg_seri)[0], msg_recv)

    for y in msg_send_list:
        do_send_recv(y)


class TestJSONObjectMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for JSONObjectMetaschemaType class."""

    _mod = 'JSONObjectMetaschemaType'
    _cls = 'JSONObjectMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestJSONObjectMetaschemaType, self).__init__(*args, **kwargs)
        self._value = {}
        self._fulldef = {'type': self.import_cls.name,
                         'properties': {}}
        self._typedef = {'properties': {}}
        for i, k in zip(range(container_utils._count), 'abcdefg'):
            self._value[k] = container_utils._vallist[i]
            self._fulldef['properties'][k] = container_utils._deflist[i]
            self._typedef['properties'][k] = container_utils._typedef[i]
        self._valid_encoded = [self._fulldef]
        self._valid_decoded = [self._value]
        self._invalid_encoded += [
            {'type': self._fulldef['type'],
             'properties': {'a': self._fulldef['properties']['a']}}]
        self._invalid_encoded.append(copy.deepcopy(self._fulldef))
        del self._invalid_encoded[-1]['properties']['a']['type']
        self._invalid_encoded.append(copy.deepcopy(self._fulldef))
        self._invalid_encoded[-1]['properties']['a']['type'] = 'invalid'
        self._compatible_objects = [(self._value, self._value, None)]
